#!/usr/bin/env python
"""
TOF bin export launcher
"""

import json
import logging
import os
import warnings

import numpy as np
from NeuNorm.normalization import Normalization
from qtpy.QtWidgets import QApplication, QDialog, QFileDialog

from ibeatles import DataType, load_ui
from ibeatles.core.detector import compute_counts_with_uncertainty
from ibeatles.session import SessionSubKeys
from ibeatles.tools.tof_bin.event_handler import EventHandler as TofBinEventHandler
from ibeatles.tools.tof_bin.utilities.get import Get
from ibeatles.tools.tof_bin.utilities.time_spectra import export_time_stamp_file
from ibeatles.tools.tof_bin.utilities.units import (
    DistanceUnitOptions,
    EnergyUnitOptions,
    TimeUnitOptions,
    convert_array_from_time_to_energy,
)
from ibeatles.tools.utilities import CombineAlgorithm, TimeSpectraKeys
from ibeatles.tools.utilities.reload.reload import Reload
from ibeatles.utilities.file_handler import FileHandler
from ibeatles.utilities.status_message_config import (
    StatusMessageStatus,
    show_status_message,
)

warnings.filterwarnings("ignore")

# label for combobox
NONE = "None"
FULL_IMAGE_LABEL = "Full image"
ROI_LABEL = "Images of ROI selected"


class ExportDataType:
    """Type of image to export, either the full image or only the roi selected"""

    full_image = "full_image"
    roi = "roi"


class TofBinExportLauncher(QDialog):
    def __init__(self, parent=None, top_parent=None):
        self.parent = parent
        self.top_parent = top_parent
        QDialog.__init__(self, parent=parent)
        self.ui = load_ui("ui_tof_bin_export.ui", baseinstance=self)
        self.update_buttons()

    def update_buttons(self):
        self.ui.reload_full_image_checkBox.setEnabled(self.ui.rebin_full_images_checkBox.isChecked())
        self.ui.distance_source_detector_label.setText("d<sub> source-detector</sub>")
        self.ui.distance_source_detector_value.setText(self.top_parent.ui.distance_source_detector.text())
        self.ui.detector_offset_units.setText("\u00b5s")
        self.ui.detector_offset_value.setText(self.top_parent.ui.detector_offset.text())

    def rebin_full_image_checkBox_clicked(self):
        self.update_buttons()

    def bin_and_export_radio_button_clicked(self):
        pass

    def bin_and_export(self, output_folder=None, data_type=ExportDataType.full_image):
        logging.info(f"binning and exporting {data_type}:")

        FileHandler.make_or_reset_folder(output_folder)
        logging.info(f" -> exported to {output_folder}")

        o_get = Get(parent=self.parent)
        bins_dict = o_get.current_bins_activated()
        number_of_bins = len(bins_dict[TimeSpectraKeys.file_index_array])

        file_index_array = bins_dict[TimeSpectraKeys.file_index_array]
        tof_array = bins_dict[TimeSpectraKeys.tof_array]
        lambda_array = bins_dict[TimeSpectraKeys.lambda_array]

        # initialize progress bar
        self.parent.eventProgress.setMinimum(0)
        self.parent.eventProgress.setMaximum(number_of_bins - 1)
        self.parent.eventProgress.setValue(0)
        self.parent.eventProgress.setVisible(True)

        file_info_dict = {}
        number_of_files_created = 0

        counts_array = []

        for _index, _bin in enumerate(file_index_array):
            logging.info(f" working with bin#{_index}")

            if len(_bin) == 0:
                logging.info(" -> empty bin, skipping!")
                self.parent.eventProgress.setValue(_index + 1)
                continue

            short_file_name = f"image_{_index:04d}.tif"
            output_file_name = os.path.join(output_folder, short_file_name)

            # create array of that bin
            _image = self.extract_data_for_this_bin(
                list_runs=_bin,
                full_image_flag=(data_type == ExportDataType.full_image),
            )
            _image = self.extract_data_for_this_bin(
                list_runs=_bin,
                full_image_flag=(data_type == ExportDataType.full_image),
            )

            # full image export
            counts_array.append(int(np.sum(_image)))
            o_norm = Normalization()
            o_norm.load(data=_image)
            o_norm.data["sample"]["file_name"][0] = os.path.basename(output_file_name)
            o_norm.export(folder=output_folder, data_type="sample", file_type="tiff")
            logging.info(f" -> exported {output_file_name}")

            file_info_dict[short_file_name] = {
                "file_index": _bin,
                "tof": tof_array[_index],
                "lambda": lambda_array[_index],
            }
            number_of_files_created += 1
            self.parent.eventProgress.setValue(_index + 1)
            QApplication.processEvents()

        # export the json file
        metadata_file_name = os.path.join(output_folder, "metadata.json")
        with open(metadata_file_name, "w") as json_file:
            json.dump(file_info_dict, json_file)

        self.parent.eventProgress.setVisible(False)
        QApplication.processEvents()

        # export the new time stamp file
        export_time_stamp_file(
            counts_array=counts_array,
            tof_array=self.parent.time_spectra[TimeSpectraKeys.tof_array],
            file_index_array=file_index_array,
            export_folder=output_folder,
        )

    def ok_clicked(self):
        logging.info("User clicked OK to export binned images")

        # check first if we want with experimental errors or not
        if self.ui.use_experimental_errors_radioButton.isChecked():
            logging.info("\tUser wants to use experimental errors")
            shutter_counts_file = self.parent.shutter_counts_file
            logging.info(f"\tShutter counts file used: {shutter_counts_file}")

            if shutter_counts_file is None or not os.path.exists(shutter_counts_file):
                logging.info(f"Shutter counts file not found: {shutter_counts_file}")
                logging.info("User will browse for the shutter count file")
                o_event = TofBinEventHandler(parent=self.parent, top_parent=self.top_parent)
                o_event.browse_for_shutter_counts_file()

        self.ok_clicked_step2()

    # def ok_clicked_step2(self):
    #     logging.info("User clicked OK to export binned images")

    #     # check first if we want with experimental errors or not
    #     if self.ui.use_experimental_errors_radioButton.isChecked():
    #         logging.info("\tUser wants to use experimental errors")
    #         shutter_counts_file = self.parent.shutter_counts_file
    #         logging.info(f"\tShutter counts file used: {shutter_counts_file}")

    #         if shutter_counts_file is None or not os.path.exists(shutter_counts_file):
    #             logging.info(f"Shutter counts file not found: {shutter_counts_file}")
    #             logging.info("User will browse for the shutter count file")
    #             o_event = TofBinEventHandler(parent=self.parent, top_parent=self.top_parent)
    #             o_event.browse_for_shutter_counts_file()

    #     self.ok_clicked_step2()

    def ok_clicked_step2(self):
        working_dir = self.top_parent.session_dict[DataType.sample][SessionSubKeys.current_folder]

        message = "Working on exporting files ..."
        show_status_message(
            parent=self.parent,
            message=message,
            status=StatusMessageStatus.working,
        )

        message = "Working on exporting files ..."
        show_status_message(
            parent=self.parent,
            message=message,
            status=StatusMessageStatus.working,
        )

        _folder = str(
            QFileDialog.getExistingDirectory(
                caption="Select Folder to export binned Images",
                directory=working_dir,
                options=QFileDialog.ShowDirsOnly,
            )
        )

        if _folder == "":
            logging.info("User cancel export binned images!")
            self.close()
            return

        self.close()

        # define output folder names
        base_folder_name = os.path.basename(os.path.dirname(self.parent.list_tif_files[0]))
        time_stamp = FileHandler.get_current_timestamp()

        nbr_output_created_dict = {"images": 0, "ascii": 0}

        nbr_output_created_dict = {"images": 0, "ascii": 0}

        output_folder_full_image = ""
        if self.ui.rebin_full_images_checkBox.isChecked():
            output_folder_full_image = os.path.join(_folder, f"{base_folder_name}_full_image_binned_{time_stamp}")
            self.bin_and_export(
                output_folder=output_folder_full_image,
                data_type=ExportDataType.full_image,
            )
            nbr_output_created_dict["images"] += 1
            nbr_output_created_dict["images"] += 1

        output_folder_roi = ""
        if self.ui.rebin_roi_selected_checkBox.isChecked():
            output_folder_roi = os.path.join(_folder, f"{base_folder_name}_roi_binned_{time_stamp}")
            self.bin_and_export(
                output_folder=output_folder_roi,
                data_type=ExportDataType.roi,
            )
            nbr_output_created_dict["images"] += 1

        if self.ui.ascii_full_image_checkBox.isChecked():
            output_file_name = os.path.join(_folder, f"{base_folder_name}_full_image_binned_{time_stamp}.txt")
            self.export_ascii(
                add_uncertainties=self.ui.use_experimental_errors_radioButton.isChecked(),
                data_type=ExportDataType.full_image,
                output_file_name=output_file_name,
            )
            nbr_output_created_dict["ascii"] += 1

        if self.ui.ascii_roi_selected_checkBox.isChecked():
            output_file_name = os.path.join(_folder, f"{base_folder_name}_roi_binned_{time_stamp}.txt")
            self.export_ascii(
                add_uncertainties=self.ui.use_experimental_errors_radioButton.isChecked(),
                data_type=ExportDataType.roi,
                output_file_name=output_file_name,
            )
            nbr_output_created_dict["ascii"] += 1
            # self.bin_and_export(
            #     output_folder=output_folder_roi,
            #     data_type=ExportDataType.roi,
            # )
            # nbr_output_created_dict["images"] += 1

        if self.ui.ascii_full_image_checkBox.isChecked():
            output_file_name = os.path.join(_folder, f"{base_folder_name}_full_image_binned_{time_stamp}.txt")
            self.export_ascii(
                add_uncertainties=self.ui.use_experimental_errors_radioButton.isChecked(),
                data_type=ExportDataType.full_image,
                output_file_name=output_file_name,
            )
            nbr_output_created_dict["ascii"] += 1

        if self.ui.ascii_roi_selected_checkBox.isChecked():
            output_file_name = os.path.join(_folder, f"{base_folder_name}_roi_binned_{time_stamp}.txt")
            self.export_ascii(
                add_uncertainties=self.ui.use_experimental_errors_radioButton.isChecked(),
                data_type=ExportDataType.roi,
                output_file_name=output_file_name,
            )
            nbr_output_created_dict["ascii"] += 1

        # reload if user requested it
        want_to_reload = (
            self.ui.reload_full_image_checkBox.isEnabled() and self.ui.reload_full_image_checkBox.isChecked()
        )
        if not want_to_reload:
            logging.info("Nothing to reload, we are done here!")
            message = (
                f"{nbr_output_created_dict['images']} binned images and {nbr_output_created_dict['ascii']} "
                + "ASCII files created successfully!"
            )
            show_status_message(
                parent=self.parent,
                message=message,
                status=StatusMessageStatus.ready,
                duration_s=15,
            )

        else:
            if os.path.exists(output_folder_full_image):
                o_reload = Reload(parent=self.parent, top_parent=self.top_parent)
                o_reload.run(data_type=DataType.normalized, output_folder=output_folder_full_image)
                self.parent.close()

    def compute_counts_array_only(self, tiff_stack=None, roi_mask=None):
        """Compute counts array for a given tiff stack and roi mask.

        :param tiff_stack: The tiff stack to process.
        :param roi_mask: The roi mask to apply.
        """
        if roi_mask is not None:
            _tiff_stack = [tiff[roi_mask] for tiff in tiff_stack]
        else:
            _tiff_stack = tiff_stack

        counts_array = np.sum(_tiff_stack, axis=1)
        return counts_array

    def export_ascii(self, add_uncertainties=False, data_type=ExportDataType.full_image, output_file_name=None):
        """Export TOF binned data to an ASCII text file.

        The exported file contains, for each active bin: bin index, mean time-of-flight,
        mean wavelength, mean energy, total counts, and the associated uncertainty.

        :param add_uncertainties: If True, compute and include experimental uncertainties
            using shutter counts data. If False, uncertainties are set to NaN.
        :param data_type: Type of data to export. Use ExportDataType.full_image for
            full detector image or ExportDataType.roi for the selected ROI.
        :param output_file_name: Path to the ASCII file to create.
        """

        if data_type == ExportDataType.roi:
            bin_roi = self.parent.bin_roi
            x0 = bin_roi["x0"]
            y0 = bin_roi["y0"]
            width = bin_roi["width"]
            height = bin_roi["height"]
            roi_mask = np.zeros(self.parent.images_array[0].shape, dtype=bool)
            roi_mask[y0 : y0 + height, x0 : x0 + width] = True
            logging.info(f"Exporting ASCII for ROI: x0={x0}, y0={y0}, width={width}, height={height}")

        else:
            roi_mask = None
            logging.info("Exporting ASCII for full image")

        metadata_lines = ["# bin index\tmean_tof (micros)\tmean_lambda (Angstroms)\tmean_energy (eV)\ttotal_counts"]
        if add_uncertainties:
            logging.info("Adding experimental uncertainties to the exported ASCII file")
            # calculate uncertainties of all images
            result_full = compute_counts_with_uncertainty(
                tiff_stack=self.parent.images_array,
                roi_mask=roi_mask,
                tof_array=self.parent.time_spectra[TimeSpectraKeys.tof_array],
                shutter_counts_file=self.parent.shutter_counts_file,
            )

            logging.info(f"result_full: {result_full}")
            logging.info(f"result_full keys: {list(result_full.keys())}")
            counts_array = list(result_full["counts"])
            uncertainties_array = list(result_full["uncertainty"])
            metadata_lines[0] += "\tuncertainty"

        else:
            counts_array = self.compute_counts_array_only(tiff_stack=self.parent.images_array, roi_mask=roi_mask)

        # format of the output ascii file
        # index, start tof, end tof, start lambda, end lambda, start energy, end energy, total counts, uncertainty
        o_get = Get(parent=self.parent)
        bins_dict = o_get.current_bins_activated()
        file_index_array = bins_dict[TimeSpectraKeys.file_index_array]
        logging.info(f"{file_index_array = }")
        statistics_dict = self.parent.statistics_dict

        logging.info(f"\t{statistics_dict = }")

        data_lines = []

        for _bin_index in statistics_dict.keys():
            list_file_index = statistics_dict[_bin_index]["list_file"]
            logging.info(f"\tProcessing bin index {_bin_index} with file indices {list_file_index}")

            if list_file_index is None:
                logging.info(f"\tSkipping bin index {_bin_index} as it has no files.")
                continue

            list_total_counts = [counts_array[_file_index] for _file_index in list_file_index]
            logging.info(f"\t{list_total_counts = }")
            mean_counts = np.mean(list_total_counts)
            logging.info(f"\t{mean_counts = }")

            if add_uncertainties:
                _uncertainties_array = [uncertainties_array[_file_index] for _file_index in list_file_index]
                uncertainty = self.calculate_uncertainty_of_mean(_uncertainties_array)

            list_lambda = statistics_dict[_bin_index]["list_lambda"]
            average_lambda = np.mean(list_lambda)
            # average_energy_ev = convert_from_wavelength_to_energy_ev(average_lambda)

            list_tof = statistics_dict[_bin_index]["list_tof"]
            average_tof = np.mean(list_tof)

            distance_source_detector_m = float(self.ui.distance_source_detector_value.text())
            detector_offset = float(self.ui.detector_offset_value.text())

            average_energy_ev = convert_array_from_time_to_energy(
                time_array=np.array([average_tof]),
                time_unit=TimeUnitOptions.us,
                distance_source_detector=distance_source_detector_m,
                distance_source_detector_unit=DistanceUnitOptions.m,
                detector_offset=detector_offset,
                detector_offset_unit=TimeUnitOptions.us,
                energy_unit=EnergyUnitOptions.eV,
            )[0]

            if add_uncertainties:
                data_lines.append(
                    f"{_bin_index}\t{average_tof:.6f}\t{average_lambda:.6f}\t{average_energy_ev:.6f}\t{mean_counts:.2f}\t{uncertainty:.2f}"
                )
            else:
                data_lines.append(
                    f"{_bin_index}\t{average_tof:.6f}\t{average_lambda:.6f}\t{average_energy_ev:.6f}\t{mean_counts:.2f}"
                )

        # write the output ascii file
        logging.info(f"Writing ASCII file to {output_file_name}")
        with open(output_file_name, "w") as fout:
            for _line in metadata_lines:
                fout.write(f"{_line}\n")
            for _line in data_lines:
                fout.write(f"{_line}\n")

    def calculate_uncertainty_of_mean(self, list_uncertainties):
        """calculate the uncertainty of the mean of counts
        assuming the uncertainties are independent

        :param list_uncertainties: list of uncertainties to mean
        :return:
            uncertainty of the mean
        """
        n = len(list_uncertainties)

        sum_squared = 0.0
        for _uncertainty in list_uncertainties:
            sum_squared += _uncertainty**2

        return np.sqrt(sum_squared) / np.sqrt(n)

    def extract_data_for_this_bin(self, list_runs=None, full_image_flag=True):
        """
        this method isolate the data of only the runs of the corresponding runs if full_image is True,
        otherwise will return the ROI selected

        :param
            list_runs: list of runs to extract
            full_image_flag: True or False (False if we want only the ROI selected)
        :return:
            image binned
        """

        data_to_work_with = []
        for _run_index in list_runs:
            data_to_work_with.append(self.parent.images_array[_run_index])

        if not full_image_flag:
            bin_roi = self.parent.bin_roi
            x0 = bin_roi["x0"]
            y0 = bin_roi["y0"]
            width = bin_roi["width"]
            height = bin_roi["height"]
            region_to_work_with = [_data[y0 : y0 + height, x0 : x0 + width] for _data in data_to_work_with]
            data_to_work_with = region_to_work_with

        # how to add images
        o_get = Get(parent=self.parent)
        bin_method = o_get.bin_add_method()
        if bin_method == CombineAlgorithm.mean:
            image_to_export = np.mean(data_to_work_with, axis=0)
        elif bin_method == CombineAlgorithm.median:
            image_to_export = np.median(data_to_work_with, axis=0)
        else:
            raise NotImplementedError("this method of adding the binned images is not supported!")

        return image_to_export



#!/usr/bin/env python
"""Core functions for strain mapping from Bragg edge fitting results."""

from typing import Dict, Optional

from lmfit.model import ModelResult


def calculate_strain_mapping(
    fit_results: Dict[str, Optional[ModelResult]],
    d0: float,
    quality_threshold: float = 0.8,
) -> Dict[str, Dict[str, float]]:
    """Calculate strain from Bragg edge fitting results.

    Parameters
    ----------
    fit_results : Dict[str, Optional[ModelResult]]
        Dictionary of fitting results for each bin
    d0 : float
        Reference d-spacing value (unstrained)
    quality_threshold : float
        R-squared threshold for accepting fit results

    Returns
    -------
    Dict[str, Dict[str, float]]
        Dictionary containing strain results for each bin:
        {bin_id: {'strain': value, 'error': value, 'quality': value}}
    """
    strain_results = {}

    for bin_id, fit_result in fit_results.items():
        # skip failed fits
        if fit_result is None:
            continue

        # skip low quality fits
        if fit_result.rsquared < quality_threshold:
            continue

        # Calculate strain: (d - d0)/d0 where d = λ/2
        wavelength = fit_result.best_values["bragg_edge_wavelength"]
        wavelength_error = fit_result.params["bragg_edge_wavelength"].stderr

        d_spacing = wavelength / 2.0
        strain = (d_spacing - d0) / d0

        # Calculate error if possible
        strain_error = wavelength_error / (2.0 * d0) if wavelength_error else None

        strain_results[bin_id] = {
            "strain": strain,
            "error": strain_error,
            "quality": fit_result.rsquared,
        }

    return strain_results



#!/usr/bin/env python
"""Module entry"""

import argparse
import sys
from pathlib import Path

from ibeatles.app.cli import main as cli_main
from ibeatles.ibeatles import main as gui_main


def main():
    parser = argparse.ArgumentParser(description="iBeatles - Bragg Edge Analysis Tool")
    parser.add_argument("--no-gui", action="store_true", help="Run in CLI mode")
    parser.add_argument("config", nargs="?", help="Path to the configuration file for CLI mode")
    parser.add_argument("--log", type=Path, help="Path to the log file (optional)")

    args = parser.parse_args()

    if args.no_gui:
        if not args.config:
            print("Error: Configuration file is required for CLI mode.")
            sys.exit(1)
        cli_main(args.config, args.log)
    else:
        gui_main(sys.argv)


if __name__ == "__main__":
    sys.exit(main())



import os

from qtpy.uic import loadUi

from ibeatles import ui

try:
    from ._version import __version__  # noqa: F401
except ImportError:
    __version__ = "unknown"

root = os.path.dirname(os.path.realpath(__file__))
refresh_image = os.path.join(root, "icons/refresh.png")
up_image = os.path.join(root, "icons/up_arrow.png")
down_image = os.path.join(root, "icons/down_arrow.png")
settings_image = os.path.join(root, "icons/plotSettings.png")
vertical_splitter_file = os.path.join(root, "icons/vertical_splitter_handle.png")
horizontal_splitter_file = os.path.join(root, "icons/horizontal_splitter_handle.png")
infos_file = os.path.join(root, "icons/info.png")
preview_file = os.path.join(root, "icons/preview.png")
error_icon_file = os.path.join(root, "icons/error_icon.png")

# tools
fitting_image = os.path.join(root, "icons/fitting_logo.png")
pixel_binning_image = os.path.join(root, "icons/pixel_binning_logo.png")
rotate_image = os.path.join(root, "icons/rotate_logo.png")
strain_mapping_image = os.path.join(root, "icons/strain_mapping_logo.png")
tof_binning_image = os.path.join(root, "icons/tof_binning_logo.png")
tof_combine_image = os.path.join(root, "icons/combine_tof_logo.png")
right_blue_arrow = os.path.join(root, "icons/right_blue_arrow.png")

# tof bin
bin_image = os.path.join(root, "icons/bin.png")
auto_image = os.path.join(root, "icons/auto.png")
manual_image = os.path.join(root, "icons/manual.png")
more_infos_image = os.path.join(root, "icons/more_infos.png")
stats_table_image = os.path.join(root, "icons/stats_table.png")
stats_plot_image = os.path.join(root, "icons/stats_plot.png")

# steps
step1_icon = os.path.join(root, "static/step1.png")
step2_icon = os.path.join(root, "static/step2.png")
step3_icon = os.path.join(root, "static/step3.png")
step4_icon = os.path.join(root, "static/step4.png")

DEFAULT_ROI = ["default", "0", "0", "20", "20", "0"]
DEFAULT_NORMALIZATION_ROI = [True, "0", "0", "20", "20", "background"]
DEFAULT_BIN = [0, 0, 20, 20, 10]
BINNING_LINE_COLOR = (255, 0, 0, 255, 1.0)

interact_me_style = "background-color: lime"
error_style = "background-color: red"
normal_style = ""

ANGSTROMS = "\u212b"
LAMBDA = "\u03bb"
MICRO = "\u00b5"
SUB_0 = "\u2080"
DELTA = "\u0394"

MATERIAL_BRAGG_PEAK_TO_DISPLAY_AT_THE_SAME_TIME = 4


class XAxisMode:
    tof_mode = "tof"
    lambda_mode = "lambda"
    file_index_mode = "file_index"


class DataType:
    sample = "sample"
    ob = "ob"
    df = "df"
    normalized = "normalized"
    normalization = "normalization"
    bin = "bin"
    fitting = "fitting"
    time_spectra = "time_spectra"
    time_spectra_normalized = "time_spectra_normalized"
    none = "none"


class ScrollBarParameters:
    maximum = "maximumx value"
    value = "current value"


class RegionType:
    sample = "sample"
    background = "background"


def load_ui(ui_filename, baseinstance):
    ui_filename = os.path.split(ui_filename)[-1]
    ui_path = os.path.dirname(ui.__file__)

    # get the location of the ui directory
    # this function assumes that all ui files are there
    filename = os.path.join(ui_path, ui_filename)
    print(f"loading {filename}")
    return loadUi(filename, baseinstance=baseinstance)


class Material:
    """
    keys to use when retrieving or saving the h, k, l, lambda and d0 value, as well as the
    lattice, crystal structure and if the user defined or not that entry.
    """

    element_name = "name of the element"
    lattice = "lattice"
    crystal_structure = "structure of the crystal"
    hkl_d0 = "list of h,k and l and respective d0 value"
    user_defined = "is this user defined?"
    via_lattice_and_crystal_structure = "using method 1 of add_element"
    via_d0 = "using method 2"
    method_used = "either via_lattice or via_d0"
    user_defined_bragg_edge_list = "list of hkl, d0 of the user defined material"
    full_list_of_element_names = "list of names of element displayed in the 2 comboBoxes"
    local_bragg_edge_list = "local Bragg edge list of material defined byt he user"


class FileType:
    ascii = "txt"
    json = "json"
    hdf5 = "h5"



#!/usr/bin/env python
"""
Plot (step 1)
"""

import numpy as np
import pyqtgraph as pg
from neutronbraggedge.experiment_handler.experiment import Experiment
from qtpy.QtGui import QBrush

from ibeatles import (
    MATERIAL_BRAGG_PEAK_TO_DISPLAY_AT_THE_SAME_TIME,
    DataType,
    ScrollBarParameters,
)
from ibeatles.binning.binning_handler import BinningHandler
from ibeatles.fitting.fitting_handler import FittingHandler
from ibeatles.utilities.colors import pen_color, roi_group_color
from ibeatles.utilities.gui_handler import GuiHandler
from ibeatles.utilities.pyqrgraph import Pyqtgrah as PyqtgraphUtilities


class CustomAxis(pg.AxisItem):
    def __init__(self, gui_parent, *args, **kwargs):
        pg.AxisItem.__init__(self, *args, **kwargs)
        self.parent = gui_parent

    def tickStrings(self, values, scale, spacing):
        strings = []

        _distance_source_detector = float(str(self.parent.ui.distance_source_detector.text()))
        _detector_offset_micros = float(str(self.parent.ui.detector_offset.text()))

        tof_s = [float(time) * 1e-6 for time in values]

        _exp = Experiment(
            tof=tof_s,
            distance_source_detector_m=_distance_source_detector,
            detector_offset_micros=_detector_offset_micros,
        )
        lambda_array = _exp.lambda_array

        for _lambda in lambda_array:
            strings.append("{:.4f}".format(_lambda * 1e10))

        return strings


class Step1Plot(object):
    data = []

    plot_ui = {
        DataType.sample: None,
        DataType.ob: None,
        DataType.normalized: None,
        "binning": None,
    }

    def __init__(self, parent=None, data_type=None, data=[]):
        self.parent = parent

        if data_type is None:
            o_gui = GuiHandler(parent=parent)
            data_type = o_gui.get_active_tab()
        self.data_type = data_type

        if len(data) == 0:
            data = self.parent.data_metadata[data_type]["data"]
        self.data = data

        self.plot_ui[DataType.sample] = self.parent.ui.bragg_edge_plot
        self.plot_ui[DataType.ob] = self.parent.ui.ob_bragg_edge_plot
        self.plot_ui[DataType.normalized] = self.parent.ui.normalized_bragg_edge_plot

    def all_plots(self):
        self.display_image()
        self.display_bragg_edge()

    def display_image(self, add_mean_radio_button_changed=False):
        """
        display the top right images + ROI boxes
        """

        _data = self.data
        _state = None
        self.parent.live_data = _data

        if len(_data) == 0:
            self.clear_plots(data_type=self.data_type)

        else:
            _data = np.array(_data)
            if self.data_type == DataType.sample:
                o_pyqt = PyqtgraphUtilities(
                    parent=self.parent,
                    image_view=self.parent.ui.image_view,
                    data_type=self.data_type,
                    add_mean_radio_button_changed=add_mean_radio_button_changed,
                )
                _state = o_pyqt.get_state()
                o_pyqt.save_histogram_level()

                self.parent.ui.area.setVisible(True)
                self.parent.ui.image_view.setImage(_data)
                self.add_origin_label(self.parent.ui.image_view)

                o_pyqt.set_state(_state)
                o_pyqt.reload_histogram_level()

            elif self.data_type == DataType.ob:
                o_pyqt = PyqtgraphUtilities(
                    parent=self.parent,
                    image_view=self.parent.ui.ob_image_view,
                    data_type=self.data_type,
                    add_mean_radio_button_changed=add_mean_radio_button_changed,
                )
                _state = o_pyqt.get_state()
                o_pyqt.save_histogram_level()
                self.parent.ui.ob_area.setVisible(True)
                self.parent.ui.ob_image_view.setImage(_data)
                self.add_origin_label(self.parent.ui.ob_image_view)
                o_pyqt.set_state(_state)
                o_pyqt.reload_histogram_level()

            elif self.data_type == DataType.normalized:
                o_pyqt = PyqtgraphUtilities(
                    parent=self.parent,
                    image_view=self.parent.ui.normalized_image_view,
                    data_type=self.data_type,
                    add_mean_radio_button_changed=add_mean_radio_button_changed,
                )
                _state = o_pyqt.get_state()
                o_pyqt.save_histogram_level()
                self.parent.ui.normalized_area.setVisible(True)
                self.parent.ui.normalized_image_view.setImage(_data)
                self.add_origin_label(self.parent.ui.normalized_image_view)
                self.parent.data_metadata[DataType.normalized]["data_live_selection"] = _data
                o_pyqt.set_state(_state)
                o_pyqt.reload_histogram_level()

                # make sure that if we have the fitting window open, we have also at least the binning
                if self.parent.fitting_ui is not None and (self.parent.binning_ui is None):
                    self.parent.menu_view_binning_clicked()

                if self.parent.binning_ui is not None:
                    o_binning = BinningHandler(parent=self.parent)
                    o_binning.display_image(data=_data)
                    self.parent.binning_ui.ui.groupBox.setEnabled(True)
                    self.parent.binning_ui.ui.groupBox_2.setEnabled(True)
                    self.parent.binning_ui.ui.left_widget.setVisible(True)
                if self.parent.fitting_ui is not None:
                    o_fitting = FittingHandler(parent=self.parent)
                    o_fitting.display_image(data=_data)
                    o_fitting.display_roi()
                    self.parent.fitting_ui.ui.area.setVisible(True)
                    o_fitting.fill_table()
                if self.parent.rotate_ui is not None:
                    o_rotate = self.parent.rotate_ui
                    o_rotate.display_rotated_images()

            self.parent.image_view_settings[self.data_type]["state"] = _state

    def initialize_default_roi(self):
        if self.data_type == DataType.sample:
            self.add_origin_roi(self.parent.ui.image_view, self.parent.ui.image_view_roi)
        elif self.data_type == DataType.ob:
            self.add_origin_roi(self.parent.ui.ob_image_view, self.parent.ui.ob_image_view_roi)
        elif self.data_type == DataType.normalized:
            self.add_origin_roi(
                self.parent.ui.normalized_image_view,
                self.parent.ui.normalized_image_view_roi,
            )

    def add_origin_roi(self, image_view, roi_id):
        image_view.addItem(roi_id)
        self.parent.list_roi_id[self.data_type] = [roi_id]

    def add_origin_label(self, image_ui):
        # origin label
        text_id = pg.TextItem(html="<span style='color: yellow;'>(0,0)", anchor=(1, 1))
        image_ui.addItem(text_id)
        text_id.setPos(-5, -5)

        # x and y arrows directions
        y_arrow = pg.ArrowItem(
            angle=-90,
            tipAngle=35,
            baseAngle=0,
            headLen=20,
            tailLen=40,
            tailWidth=2,
            pen="y",
            brush=None,
        )
        image_ui.addItem(y_arrow)
        y_arrow.setPos(0, 65)
        y_text = pg.TextItem(html="<span style='color: yellow;'>Y")
        image_ui.addItem(y_text)
        y_text.setPos(-30, 20)

        x_arrow = pg.ArrowItem(
            angle=180,
            tipAngle=35,
            baseAngle=0,
            headLen=20,
            tailLen=40,
            tailWidth=2,
            pen="y",
            brush=None,
        )
        image_ui.addItem(x_arrow)
        x_arrow.setPos(65, 0)
        x_text = pg.TextItem(html="<span style='color: yellow;'>X")
        image_ui.addItem(x_text)
        x_text.setPos(20, -30)

    def refresh_roi(self):
        pass

    def clear_image(self, data_type="sample"):
        if data_type == "sample":
            self.parent.ui.image_view.clear()
        elif data_type == "ob":
            self.parent.ui.ob_image_view.clear()
        elif data_type == "normalized":
            self.parent.ui.normalized_image_view.clear()

    def clear_plots(self, data_type="sample"):
        if data_type == "sample":
            self.parent.ui.image_view.clear()
            self.parent.ui.bragg_edge_plot.clear()
        elif data_type == "ob":
            self.parent.ui.ob_image_view.clear()
            self.parent.ui.ob_bragg_edge_plot.clear()
        elif data_type == "normalized":
            self.parent.ui.normalized_image_view.clear()
            self.parent.ui.normalized_bragg_edge_plot.clear()

    def display_general_bragg_edge(self, data_type=None):
        if data_type is None:
            o_gui = GuiHandler(parent=self.parent)
            data_type = o_gui.get_active_tab()

        data = self.parent.data_metadata[data_type]["data"]
        self.data = data
        self.display_bragg_edge()

    def save_roi(self, label, x0, y0, x1, y1, group, data_type, index):
        _width = np.abs(x1 - x0)
        _height = np.abs(y1 - y0)

        _list_roi = self.parent.list_roi[data_type]
        if len(_list_roi) == 0:
            _label = "roi_label"
            _group = "0"
            _list_roi = [_label, str(x0), str(y0), str(_width), str(_height), _group]
            self.parent.list_roi[data_type] = [_list_roi]
        else:
            _label = label
            _group = group
            _list_roi = [_label, str(x0), str(y0), str(_width), str(_height), _group]
            self.parent.list_roi[data_type][index] = _list_roi

    def update_roi_editor(self, index):
        o_roi_editor = self.parent.roi_editor_ui[self.data_type]
        o_roi_editor.refresh(row=index)

        # o_roi = RoiHandler(parent=self.parent, data_type=self.data_type)
        # row_to_activate = o_roi.get_roi_index_that_changed()
        # o_roi_editor.activate_row(row_to_activate)

    def extract_data(self, list_data_group, data):
        list_data = {"0": [], "1": [], "2": [], "3": []}

        for _group in list_data_group.keys():
            _list_roi = list_data_group[_group]
            if len(_list_roi) == 0:
                list_data[_group] = []
            else:
                for _data in data:
                    # nbr_roi = len(_list_roi)
                    _tmp_data = []
                    for _roi in _list_roi:
                        [x0, x1, y0, y1] = _roi

                        if self.parent.ui.roi_add_button.isChecked():
                            # _tmp_data.append(np.sum(_data[y0:y1, x0:x1]))
                            _tmp_data.append(np.nansum(_data[x0:x1, y0:y1]))
                        else:
                            # _tmp_data.append(np.mean(_data[y0:y1, x0:x1]))
                            _tmp_data.append(np.nanmean(_data[x0:x1, y0:y1]))

                    if self.parent.ui.roi_add_button.isChecked():
                        list_data[_group].append(np.nansum(_tmp_data))
                    else:
                        list_data[_group].append(np.mean(_tmp_data, axis=0))

        return list_data

    def get_row_parameters(self, roi_editor_ui, row):
        # # label
        _item = roi_editor_ui.tableWidget.item(row, 0)
        if _item is None:
            raise ValueError
        label = str(_item.text())

        # x0
        _item = roi_editor_ui.tableWidget.item(row, 1)
        if _item is None:
            raise ValueError
        x0 = int(str(_item.text()))

        # y0
        _item = roi_editor_ui.tableWidget.item(row, 2)
        if _item is None:
            raise ValueError
        y0 = int(str(_item.text()))

        # width
        _item = roi_editor_ui.tableWidget.item(row, 3)
        if _item is None:
            raise ValueError
        width = int(str(_item.text()))

        # height
        _item = roi_editor_ui.tableWidget.item(row, 4)
        if _item is None:
            raise ValueError
        height = int(str(_item.text()))

        # group
        _group_widget = roi_editor_ui.tableWidget.cellWidget(row, 5)
        if _group_widget is None:
            raise ValueError
        _index_selected = _group_widget.currentIndex()
        group = str(_index_selected)

        return [label, x0, y0, width, height, group]

    def clear_bragg_edge_plot(self):
        if self.data_type == "sample":
            self.parent.ui.bragg_edge_plot.clear()
        elif self.data_type == "ob":
            self.parent.ui.ob_bragg_edge_plot.clear()
        elif self.data_type == "normalized":
            self.parent.ui.normalized_bragg_edge_plot.clear()

    def retrieve_list_data_group(self, mouse_selection=False):
        """
        this method looks at the current data_type and create a dictionary of the ROI selected
        for all the groups
        """

        list_roi_id = self.parent.list_roi_id[self.data_type]
        list_roi = self.parent.list_roi[self.data_type]

        # collect the right image_view and image_view_item to recover the ROI
        roi_editor_ui = self.parent.roi_editor_ui[self.data_type]
        if self.data_type == "sample":
            _image_view = self.parent.ui.image_view
            _image_view_item = self.parent.ui.image_view.imageItem
        elif self.data_type == "ob":
            _image_view = self.parent.ui.ob_image_view
            _image_view_item = self.parent.ui.ob_image_view.imageItem
        elif self.data_type == "normalized":
            _image_view = self.parent.ui.normalized_image_view
            _image_view_item = self.parent.ui.normalized_image_view.imageItem

        # used here to group rois into their group for Bragg Edge plot
        list_data_group = {"0": [], "1": [], "2": [], "3": []}

        for _index, roi in enumerate(list_roi_id):
            if mouse_selection:
                if isinstance(self.parent.live_data, list):
                    self.parent.live_data = np.array(self.parent.live_data)

                try:
                    region = roi.getArraySlice(self.parent.live_data, _image_view_item)
                except IndexError:
                    return

                label = list_roi[_index][0]
                x0 = region[0][0].start
                x1 = region[0][0].stop - 1
                y0 = region[0][1].start
                y1 = region[0][1].stop - 1
                group = list_roi[_index][-1]

                if x1 == x0:
                    x1 += 1
                if y1 == y0:
                    y1 += 1

            else:
                if roi_editor_ui is None:
                    [label, x0, y0, w, h, group] = list_roi[_index]
                    x0 = int(x0)
                    y0 = int(y0)
                    w = int(w)
                    h = int(h)

                else:
                    try:
                        [label, x0, y0, w, h, group] = self.get_row_parameters(roi_editor_ui.ui, _index)

                    except ValueError:
                        return

                x1 = x0 + w
                y1 = y0 + h
                roi.setPos([x0, y0], update=False, finish=False)
                roi.setSize([w, h], update=False, finish=False)

            # display ROI boxes
            roi.setPen(pen_color[group])

            _text_array = self.parent.list_label_roi_id[self.data_type]
            if len(_text_array) == 0:
                text_id = pg.TextItem(
                    html='<div style="text-align: center"><span style="color: #ff0000;">' + label + "</span></div>",
                    anchor=(-0.3, 1.3),
                    border="w",
                    fill=(0, 0, 255, 50),
                )
                _image_view.addItem(text_id)
                text_id.setPos(x0, y0)
                self.parent.list_label_roi_id[self.data_type].append(text_id)
            else:
                text_id = self.parent.list_label_roi_id[self.data_type][_index]
                # text_id.setText(label)
                text_id.setPos(x0, y0)
                text_id.setHtml(
                    '<div style="text-align: center"><span style="color: #ff0000;">' + label + " \
                                                                                              "
                    "</span></div>"
                )

            list_data_group[group].append([x0, x1, y0, y1])

            self.save_roi(label, x0, y0, x1, y1, group, self.data_type, _index)

            if mouse_selection:
                if roi_editor_ui is not None:
                    roi_editor_ui.ui.tableWidget.blockSignals(True)
                    self.update_roi_editor(_index)
                    roi_editor_ui.ui.tableWidget.blockSignals(False)

        return list_data_group

    def display_bragg_edge(self, mouse_selection=True):
        """
        Display the bottom right plot showing the bragg edges and the position of the material bragg peaks
        """
        _data = self.data
        # list_roi = self.parent.list_roi[self.data_type]

        if len(_data) == 0:  # clear data if no data
            self.clear_bragg_edge_plot()

        else:  # retrieve dictionaries of roi_id and roi data (label, x, y, w, h, group)
            list_data_group = self.retrieve_list_data_group(mouse_selection=mouse_selection)

            # work over groups
            data = self.parent.data_metadata[self.data_type]["data"]
            bragg_edges = self.extract_data(list_data_group, data)

            # check if xaxis can be in lambda, or tof

            if self.data_type in [DataType.sample, DataType.ob]:
                time_spectra_file = self.parent.data_metadata[DataType.sample]["time_spectra"]["filename"]
            else:
                time_spectra_file = self.parent.data_metadata[self.data_type]["time_spectra"]["filename"]
            o_gui = GuiHandler(parent=self.parent)

            if time_spectra_file == "":
                o_gui.enable_xaxis_button(tof_flag=False)
                tof_array = []
                lambda_array = []

            else:
                o_gui.enable_xaxis_button(tof_flag=True)

                if self.data_type == "normalized":
                    tof_array = self.parent.data_metadata["time_spectra"]["normalized_data"]
                    lambda_array = self.parent.data_metadata["time_spectra"]["normalized_lambda"]
                else:
                    tof_array = self.parent.data_metadata["time_spectra"]["data"]
                    lambda_array = self.parent.data_metadata["time_spectra"]["lambda"]
                self.parent.normalized_lambda_bragg_edge_x_axis = lambda_array * 1e10

            # display of bottom bragg edge plot
            dictionary = self.plot_bragg_edge(tof_array=tof_array, lambda_array=lambda_array, bragg_edges=bragg_edges)

            x_axis = dictionary["x_axis"]
            [linear_region_left, linear_region_right] = dictionary["linear_region"]
            o_gui.xaxis_label()

            lr = pg.LinearRegionItem([linear_region_left, linear_region_right])
            lr.setZValue(-10)

            if self.data_type == "sample":
                self.parent.ui.bragg_edge_plot.addItem(lr)
            elif self.data_type == "ob":
                self.parent.ui.ob_bragg_edge_plot.addItem(lr)
            else:
                self.parent.ui.normalized_bragg_edge_plot.addItem(lr)
                self.parent.fitting_bragg_edge_x_axis = x_axis

            lr.sigRegionChangeFinished.connect(self.parent.bragg_edge_selection_changed)
            self.parent.list_bragg_edge_selection_id[self.data_type] = lr
            self.parent.current_bragg_edge_x_axis[self.data_type] = x_axis

    def plot_bragg_edge(self, tof_array=[], lambda_array=[], bragg_edges=[]):
        """
        plot the bragg edges
        """
        data_type = self.data_type
        plot_ui = self.plot_ui[data_type]
        plot_ui.clear()

        list_files_selected = self.parent.list_file_selected[self.data_type]
        linear_region_left_index = int(list_files_selected[0])
        linear_region_right_index = int(list_files_selected[-1])
        linear_region_left = linear_region_left_index
        linear_region_right = linear_region_right_index

        x_axis = []
        plot_ui.setLabel("left", "Total Counts")

        _symbol = "t"

        # use to check if bragg peaks scroll bar should be visible or not
        o_gui = GuiHandler(parent=self.parent)

        if len(tof_array) == 0:
            plot_ui.setLabel("bottom", "File Index")

            for _key in bragg_edges.keys():
                _bragg_edge = bragg_edges[_key]
                if len(_bragg_edge) == 0:
                    continue
                curve = plot_ui.plot(
                    _bragg_edge,
                    symbolPen=None,
                    pen=pen_color[_key],
                    symbol=_symbol,
                    symbolSize=5,
                )
                x_axis = np.arange(len(_bragg_edge))

                curvePoint = pg.CurvePoint(curve)
                plot_ui.addItem(curvePoint)
                _text = pg.TextItem("Group {}".format(_key), anchor=(0.5, 0))
                _text.setParentItem(curvePoint)
                brush = QBrush()
                brush.setColor(roi_group_color[int(_key)])
                arrow = pg.ArrowItem(angle=0, brush=brush)
                arrow.setParentItem(curvePoint)
                curvePoint.setPos(x_axis[-1])

            o_gui.update_bragg_peak_scrollbar(force_hide_widgets=True)

        else:
            tof_array = tof_array * 1e6

            xaxis_choice = o_gui.get_xaxis_checked(data_type=self.data_type)
            o_gui.update_bragg_peak_scrollbar(xaxis_mode=xaxis_choice)

            first_index = True

            for _key in bragg_edges.keys():
                _bragg_edge = bragg_edges[_key]
                if len(_bragg_edge) == 0:
                    continue

                if xaxis_choice == "file_index":
                    curve = plot_ui.plot(
                        _bragg_edge,
                        pen=pen_color[_key],
                        symbolPen=None,
                        symbolSize=5,
                        symbol=_symbol,
                    )
                    x_axis = np.arange(len(_bragg_edge))

                elif xaxis_choice == "tof":
                    curve = plot_ui.plot(
                        tof_array,
                        _bragg_edge,
                        pen=pen_color[_key],
                        symbolPen=None,
                        symbolSize=5,
                        symbol=_symbol,
                    )
                    x_axis = tof_array
                    linear_region_left = tof_array[linear_region_left_index]
                    linear_region_right = tof_array[linear_region_right_index]

                else:  # lambda
                    if first_index:
                        lambda_array = lambda_array * 1e10

                    curve = plot_ui.plot(
                        lambda_array,
                        _bragg_edge,
                        pen=pen_color[_key],
                        symbolPen=None,
                        symbolSize=5,
                    )
                    x_axis = lambda_array

                    linear_region_left = lambda_array[linear_region_left_index]
                    linear_region_right = lambda_array[linear_region_right_index]

                    if first_index:
                        self.display_selected_element_bragg_edges(
                            plot_ui=plot_ui,
                            lambda_range=[lambda_array[0], lambda_array[-1]],
                            ymax=np.nanmax(_bragg_edge),
                        )
                        first_index = False

                curvePoint = pg.CurvePoint(curve)
                plot_ui.addItem(curvePoint)
                _text = pg.TextItem("Group {}".format(_key), anchor=(0.5, 0), color=pen_color[_key])
                _text.setParentItem(curvePoint)
                brush = QBrush()
                brush.setColor(roi_group_color[int(_key)])
                arrow = pg.ArrowItem(angle=0, brush=brush)
                arrow.setParentItem(curvePoint)

                if xaxis_choice == "lambda":
                    last_position = x_axis[-1]
                else:
                    last_position = x_axis[-1]

                curvePoint.setPos(last_position)

        return {
            "x_axis": x_axis,
            "linear_region": [linear_region_left, linear_region_right],
        }

    def display_selected_element_bragg_edges(self, plot_ui=plot_ui, lambda_range=None, ymax=0):
        display_flag = self.parent.ui.material_display_checkbox.isChecked()
        if not display_flag:
            return

        _selected_element_bragg_edges_array = self.parent.selected_element_bragg_edges_array
        _selected_element_hkl_array = self.parent.selected_element_hkl_array

        # nbr_hkl_in_list = len(_selected_element_bragg_edges_array)
        # nbr_to_display_at_the_same_time = 4

        hkl_scrollbar_ui = self.parent.hkl_scrollbar_ui["widget"][self.data_type]
        hkl_scrollbar_ui.blockSignals(True)
        max_value = self.parent.hkl_scrollbar_dict[ScrollBarParameters.maximum]
        hkl_scrollbar_ui.setMaximum(max_value)
        current_value = self.parent.hkl_scrollbar_dict[ScrollBarParameters.value]
        hkl_scrollbar_ui.setValue(current_value)
        hkl_scrollbar_ui.blockSignals(False)

        list_to_display = np.arange(
            max_value - current_value,
            max_value - current_value + MATERIAL_BRAGG_PEAK_TO_DISPLAY_AT_THE_SAME_TIME,
        )

        # display only the vertical lines
        for _index, _x in enumerate(_selected_element_bragg_edges_array):
            # if (_x >= lambda_range[0]) and (_x <= lambda_range[1]):

            if _x is None:
                continue

            _x = float(_x)

            if _index in list_to_display:
                # label of line
                _hkl = _selected_element_hkl_array[_index]
                _hkl_formated = "{},{},{}".format(_hkl[0], _hkl[1], _hkl[2])
                _text = pg.TextItem(_hkl_formated, anchor=(0, 1), angle=45, color=pg.mkColor("c"))
                _text.setPos(_x, ymax)
                plot_ui.addItem(_text)

                # vertical line
                _item = pg.InfiniteLine(_x, pen=pg.mkPen("c"))
                plot_ui.addItem(_item)



#!/usr/bin/env python
"""Normalization functions for the TOF imaging data."""

import logging
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from NeuNorm.normalization import Normalization as NeuNormNormalization
from NeuNorm.roi import ROI

from ibeatles.core.config import IBeatlesUserConfig
from ibeatles.core.processing.moving_average import moving_average


def normalize_data(
    sample_data: np.ndarray,
    ob_data: Optional[np.ndarray],
    time_spectra: Dict[str, Any],
    config: IBeatlesUserConfig,
    output_folder: str,
) -> Tuple[np.ndarray, str]:
    """
    Normalize the input data based on the provided configuration.

    Parameters:
    -----------
    sample_data : np.ndarray
        The raw sample data to be normalized.
    ob_data : Optional[np.ndarray]
        The open beam data, if available.
    time_spectra : Dict[str, Any]
        Time spectra information as returned by load_time_spectra function.
    config : IBeatlesUserConfig
        Configuration object containing normalization settings.
    output_folder : str
        Base output folder for normalized data.

    Returns:
    --------
    Tuple[np.ndarray, str]
        A tuple containing the normalized data and the full path to the output folder.
    """
    logging.info("Starting normalization process")

    # Create NeuNorm object
    o_norm = NeuNormNormalization()
    o_norm.load(data=sample_data)
    if ob_data is not None:
        o_norm.load(data=ob_data, data_type="ob")

    # Apply moving average if configured
    if config.normalization.moving_average.active:
        logging.info("Applying moving average")
        if config.normalization.processing_order == "Moving average, Normalization":
            o_norm.data["sample"]["data"] = moving_average(
                np.array(o_norm.data["sample"]["data"]),  # NeuNorm is return a list of arrays
                config.normalization.moving_average,
            )
            if ob_data is not None:
                o_norm.data["ob"]["data"] = moving_average(
                    np.array(o_norm.data["ob"]["data"]),
                    config.normalization.moving_average,
                )

    # Perform normalization
    logging.info("Performing normalization")
    background_roi = _get_background_roi(config)

    if ob_data is None:
        o_norm.normalization(roi=background_roi, use_only_sample=True)
    else:
        o_norm.normalization(roi=background_roi if background_roi else None)

    # Apply moving average after normalization if configured
    if (
        config.normalization.moving_average.active
        and config.normalization.processing_order == "Normalization, Moving Average"
    ):
        logging.info("Applying moving average after normalization")
        normalized_data = moving_average(
            np.array(o_norm.get_normalized_data()),
            config.normalization.moving_average,
        )
        # manual replace the normalized data
        o_norm.data["normalized"] = normalized_data

    # Export normalized data
    full_output_folder = _create_output_folder(output_folder, config)
    o_norm.export(folder=full_output_folder)

    # Move time spectra file
    _copy_time_spectra(time_spectra, full_output_folder)

    return o_norm.get_normalized_data(), full_output_folder


def _get_background_roi(config: IBeatlesUserConfig) -> Optional[List[ROI]]:
    """Extract background ROI from configuration."""
    if config.normalization.sample_background:
        return [
            ROI(x0=bg.x0, y0=bg.y0, width=bg.width, height=bg.height) for bg in config.normalization.sample_background
        ]
    return None


def _create_output_folder(base_folder: str, config: IBeatlesUserConfig) -> str:
    """Create and return the full path to the output folder."""
    sample_name = Path(config.raw_data.raw_data_dir).name
    full_output_folder = Path(base_folder) / f"{sample_name}_normalized"
    full_output_folder.mkdir(parents=True, exist_ok=True)
    return str(full_output_folder)


def _copy_time_spectra(time_spectra: Dict[str, Any], output_folder: str) -> None:
    """Copy time spectra file to the output folder."""
    source_file = time_spectra["filename"]
    dest_file = Path(output_folder) / time_spectra["short_filename"]
    shutil.copy(str(source_file), str(dest_file))
    logging.info(f"Copied time spectra file to {dest_file}")



import numpy as np
from astropy.io import fits
from PIL import Image


class LoadData:
    __slots__ = ["image_array", "parent", "list_of_files", "image_ext"]

    def __init__(self, parent=None, list_of_files=[], image_ext=".fits"):
        self.parent = parent
        self.list_of_files = list_of_files
        self.image_ext = image_ext

    def load(self):
        if (self.image_ext == ".tiff") or (self.image_ext == ".tif"):
            self.load_tiff()
        elif self.image_ext == ".fits":
            self.load_fits()
        else:
            raise TypeError("Image Type not supported")

    def load_tiff(self):
        _list_of_files = self.list_of_files
        _data = []
        for _file in _list_of_files:
            _image = Image.open(_file)
            _data.append(np.asarray(_image))
            _image.close()
        self.image_array = _data

    def load_fits(self):
        _list_of_files = self.list_of_files
        _data = []
        for _file in _list_of_files:
            _image = self.load_fits_file(_file)
            _data.append(_image)

        self.image_array = _data

    @staticmethod
    def load_fits_file(fits_file_name):
        file_handler = fits.open(fits_file_name, ignore_missing_end=True)
        tmp = file_handler[0].data
        try:
            if len(tmp.shape) == 3:
                tmp = tmp.reshape(tmp.shape[1:])
                file_handler.close()
            return tmp
        except OSError:
            file_handler.close()
            raise OSError("Unable to read the FITS file provided!")



#!/usr/bin/env python
"""Visualization functions for strain mapping results."""

from typing import Dict, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from lmfit.model import ModelResult
from matplotlib.axes import Axes
from matplotlib.figure import Figure


def plot_strain_map_overlay(
    strain_results: Dict[str, Dict[str, float]],
    bin_transmission: Dict[str, Dict],
    integrated_image: np.ndarray,
    colormap: str = "viridis",
    interpolation: str = "nearest",
    alpha: float = 0.5,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
) -> Tuple[Figure, Axes]:
    """Create strain map overlay on integrated image.

    Parameters
    ----------
    strain_results : Dict[str, Dict[str, float]]
        Dictionary of strain results per bin
    bin_transmission : Dict[str, Dict]
        Dictionary containing bin coordinates
    integrated_image : np.ndarray
        Background image to overlay on
    colormap : str
        Matplotlib colormap name
    interpolation : str
        Matplotlib interpolation method
    alpha : float
        Transparency of overlay
    vmin : Optional[float]
        Minimum value for strain colormap
    vmax : Optional[float]
        Maximum value for strain colormap

    Returns
    -------
    Tuple[Figure, Axes]
        Figure and axes with plot
    """
    # Create empty strain map matching integrated image shape
    strain_map = np.full_like(integrated_image, np.nan, dtype=np.float64)

    # Fill strain values into the map
    for bin_id, result in strain_results.items():
        coords = bin_transmission[bin_id]["coordinates"]
        strain_map[coords.y0 : coords.y1, coords.x0 : coords.x1] = result["strain"]

    # Create figure and axes
    fig, ax = plt.subplots(figsize=(10, 8))

    # Plot background image in grayscale
    ax.imshow(integrated_image, cmap="gray", interpolation=interpolation)

    # Plot strain map overlay
    # If vmin/vmax not provided, use data range
    if vmin is None:
        vmin = np.nanmin(strain_map)
    if vmax is None:
        vmax = np.nanmax(strain_map)

    im = ax.imshow(
        strain_map,
        cmap=colormap,
        alpha=alpha,
        interpolation=interpolation,
        vmin=vmin,
        vmax=vmax,
    )

    # Add colorbar
    fig.colorbar(im, ax=ax, label="Strain")

    # Add title and labels
    ax.set_title("Strain Map Overlay")
    ax.set_xlabel("Pixel")
    ax.set_ylabel("Pixel")

    return fig, ax


def plot_fitting_results_grid(
    fit_results: Dict[str, Optional[ModelResult]],
    bin_transmission: Dict[str, Dict],
    reference_wavelength: float,
    figsize: Tuple[int, int] = (50, 50),
) -> Tuple[Figure, np.ndarray]:
    """Create grid plot of fitting results matching binning layout.

    Parameters
    ----------
    fit_results : Dict[str, Optional[ModelResult]]
        Dictionary of fitting results for each bin
    bin_transmission : Dict[str, Dict]
        Dictionary containing bin coordinates and data
    reference_wavelength : float
        Reference Bragg edge wavelength for comparison
    figsize : Tuple[int, int]
        Figure size in inches

    Returns
    -------
    Tuple[Figure, np.ndarray]
        Figure and array of axes
    """
    # Find grid dimensions from bin coordinates
    max_row = max(int(bin_transmission[bin_id]["coordinates"].row_index) for bin_id in bin_transmission) + 1
    max_col = max(int(bin_transmission[bin_id]["coordinates"].column_index) for bin_id in bin_transmission) + 1

    # Create figure and axes grid
    fig, axes = plt.subplots(max_row, max_col, figsize=figsize)

    # Plot each bin result in its corresponding position
    for bin_id, fit_result in fit_results.items():
        coords = bin_transmission[bin_id]["coordinates"]
        row = coords.row_index
        col = coords.column_index

        ax = axes[row, col]
        ax.set_title(f"Bin {bin_id}")

        if fit_result is not None:
            # Plot fitting result
            fit_result.plot_fit(ax=ax, datafmt=".")

            # Add reference wavelength line
            ax.axvline(x=reference_wavelength, color="black", linestyle="--", label="Reference")

            # Add fitted wavelength line
            ax.axvline(
                x=fit_result.best_values["bragg_edge_wavelength"],
                color="green",
                linestyle="--",
                label="Fitted",
            )

        else:
            # Plot raw data
            wavelengths = bin_transmission[bin_id]["wavelengths"]
            wavelengths_angstrom = wavelengths * 1e10
            transmission = bin_transmission[bin_id]["transmission"]
            ax.scatter(wavelengths_angstrom, transmission, label="Raw Data", s=1)

        # set x and y labels
        ax.set_xlabel("Wavelength ($\AA$)")
        ax.set_ylabel("Transmission")

    plt.tight_layout()
    return fig, axes



#!/usr/bin/env python
"""
Selected Bins Handler
"""

import numpy as np
import pyqtgraph as pg

from ibeatles.fitting import FittingTabSelected
from ibeatles.fitting.display import Display as FittingDisplay
from ibeatles.fitting.fitting_functions import advanced_fit, basic_fit
from ibeatles.fitting.get import Get
from ibeatles.fitting.kropff.get import Get as KropffGet


class SelectedBinsHandler(object):
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def clear_all_selected_bins(self):
        list_bins = self.parent.list_bins_selected_item
        for _bin_ui in list_bins:
            self.parent.image_view.removeItem(_bin_ui)

    def clear_all_locked_bins(self):
        list_bins = self.parent.list_bins_locked_item
        for _bin_ui in list_bins:
            self.parent.image_view.removeItem(_bin_ui)

    def update_bins_selected(self):
        self.clear_all_selected_bins()
        o_get = Get(parent=self.parent)
        fitting_tab_selected = o_get.main_tab_selected()

        # March Dollase
        if fitting_tab_selected == FittingTabSelected.march_dollase:
            table_dictionary = self.grand_parent.march_table_dictionary
            list_bins_selected_item = []
            for _index in table_dictionary.keys():
                box = table_dictionary[_index]["selected_item"]
                if table_dictionary[_index]["active"]:
                    self.parent.image_view.addItem(box)
                    list_bins_selected_item.append(box)
            self.parent.list_bins_selected_item = list_bins_selected_item

        # kropff
        else:
            # only display the bin of the row selected
            o_get = KropffGet(parent=self.parent)
            row_selected = o_get.kropff_row_selected()
            table_dictionary = self.grand_parent.kropff_table_dictionary
            list_bins_selected_item = []
            for _index in row_selected:
                str_index = str(_index)
                box = table_dictionary[str_index]["selected_item"]
                self.parent.image_view.addItem(box)
                list_bins_selected_item.append(box)
            self.parent.list_bins_selected_item = list_bins_selected_item
            self.parent.update_kropff_fitting_plots()

    def update_bins_locked(self, all_flag=True):
        self.clear_all_locked_bins()
        o_get = Get(parent=self.parent)
        fitting_tab_selected = o_get.main_tab_selected()
        if fitting_tab_selected == FittingTabSelected.march_dollase:
            table_dictionary = self.grand_parent.march_table_dictionary
            list_bins_locked_item = []
            for _index in table_dictionary.keys():
                box = table_dictionary[_index]["locked_item"]
                if table_dictionary[_index]["lock"]:
                    self.parent.image_view.addItem(box)
                    list_bins_locked_item.append(box)
            self.parent.list_bins_locked_item = list_bins_locked_item
        else:
            return

    def retrieve_list_bin_selected(self, flag_name="active"):
        """this is looking at the table_dictionary and the flag of the 'active' or 'lock' key
        item to figure out if the row is checked or not"""

        list_bin_selected = []

        # if self.parent.bragg_edge_active_button_status:
        #     flag_name = 'active'
        # else:
        #     flag_name = 'lock'

        table_dictionary = self.grand_parent.march_table_dictionary
        for _index in table_dictionary:
            if table_dictionary[_index][flag_name]:
                list_bin_selected.append(_index)

        list_bin_selected.sort()
        return list_bin_selected

    def update_bragg_edge_plot(self):
        self.parent.bragg_edge_plot.clear()

        o_get = Get(parent=self.parent)
        fitting_tab_selected = o_get.main_tab_selected()
        if fitting_tab_selected == FittingTabSelected.march_dollase:
            self.update_march_dollase_bragg_edge_plot()
        else:
            self.update_kropff_bragg_edge_plot()

    def update_kropff_bragg_edge_plot(self):
        table_dictionary = self.grand_parent.kropff_table_dictionary

        x_axis = self.grand_parent.normalized_lambda_bragg_edge_x_axis
        self.parent.bragg_edge_data["x_axis"] = x_axis

        # retrieve image
        data_2d = self.grand_parent.data_metadata["normalized"]["data"]

        o_get = KropffGet(parent=self.parent)
        list_bin_selected = o_get.kropff_row_selected()

        bragg_edge_data = []
        # nbr_index_selected = len(list_bin_selected)
        for _bin_selected in list_bin_selected:
            _entry = table_dictionary[str(_bin_selected)]["bin_coordinates"]
            x0 = _entry["x0"]
            x1 = _entry["x1"]
            y0 = _entry["y0"]
            y1 = _entry["y1"]

            _data = data_2d[:, x0:x1, y0:y1]
            inter1 = np.nanmean(_data, axis=1)
            final = np.nanmean(inter1, axis=1)
            bragg_edge_data.append(final)

        bragg_edge_data = np.nanmean(bragg_edge_data, axis=0)
        # x_axis = self.grand_parent.normalized_lambda_bragg_edge_x_axis

        try:
            self.parent.bragg_edge_plot.plot(x_axis, bragg_edge_data)
        except Exception:
            return

        self.parent.bragg_edge_plot.setLabel("bottom", "\u03bb (\u212b)")
        self.parent.bragg_edge_plot.setLabel("left", "Average Counts")

        if len(self.grand_parent.fitting_bragg_edge_linear_selection) == 0:
            linear_region_left_index = 0
            linear_region_right_index = len(x_axis) - 1
            self.grand_parent.fitting_bragg_edge_linear_selection = [
                linear_region_left_index,
                linear_region_right_index,
            ]

        else:
            [linear_region_left_index, linear_region_right_index] = (
                self.grand_parent.fitting_bragg_edge_linear_selection
            )

        lr_left = x_axis[linear_region_left_index]
        lr_right = x_axis[linear_region_right_index]

        linear_region_range = [lr_left, lr_right]

        if self.parent.fitting_lr:
            self.parent.bragg_edge_plot.removeItem(self.parent.fitting_lr)

        lr = pg.LinearRegionItem(
            values=linear_region_range,
            orientation="vertical",
            brush=None,
            movable=True,
            bounds=None,
        )
        lr.setZValue(-10)
        lr.sigRegionChangeFinished.connect(self.parent.bragg_edge_linear_region_changed)
        lr.sigRegionChanged.connect(self.parent.bragg_edge_linear_region_changing)
        self.parent.bragg_edge_plot.addItem(lr)
        self.parent.fitting_lr = lr

        o_display = FittingDisplay(parent=self.parent, grand_parent=self.grand_parent)
        o_display.display_lambda_0()

    def update_march_dollase_bragg_edge_plot(self):
        if self.grand_parent.display_active_row_flag:
            flag_name = "active"
        else:
            flag_name = "lock"
        list_bin_selected = self.retrieve_list_bin_selected(flag_name=flag_name)

        if len(list_bin_selected) == 0:
            return

        table_dictionary = self.grand_parent.march_table_dictionary

        # retrieve image
        data_2d = self.grand_parent.data_metadata["normalized"]["data"]

        # isolate data selected    data[x0:x1, y0:y1] for each bin selected
        bragg_edge_data = []
        # nbr_index_selected = len(list_bin_selected)
        for _bin_selected in list_bin_selected:
            _entry = table_dictionary[str(_bin_selected)]["bin_coordinates"]
            x0 = _entry["x0"]
            x1 = _entry["x1"]
            y0 = _entry["y0"]
            y1 = _entry["y1"]
            _data = data_2d[:, x0:x1, y0:y1]
            # inter1 = np.sum(_data, axis=1)
            # final = np.sum(inter1, axis=1)
            inter1 = np.nanmean(_data, axis=1)
            final = np.nanmean(inter1, axis=1)
            bragg_edge_data.append(final)
            # if bragg_edge_data == []:
            # bragg_edge_data = final
            # else:
            # bragg_edge_data += final

        bragg_edge_data = np.nanmean(bragg_edge_data, axis=0)
        x_axis = self.grand_parent.normalized_lambda_bragg_edge_x_axis

        # save x and y-axis of bragg edge plot for initialization of a1, a2, a5 and a6
        self.parent.bragg_edge_data["x_axis"] = x_axis
        self.parent.bragg_edge_data["y_axis"] = bragg_edge_data

        self.parent.bragg_edge_plot.plot(x_axis, bragg_edge_data)
        # if self.parent.xaxis_button_ui['normalized']['file_index'].isChecked():
        # self.parent.fitting_ui.bragg_edge_plot.setLabel("bottom", "File Index")
        # elif self.parent.xaxis_button_ui['normalized']['tof'].isChecked():
        # self.parent.fitting_ui.bragg_edge_plot.setLabel("bottom", u"TOF (\u00B5s)")
        # else:
        self.parent.bragg_edge_plot.setLabel("bottom", "\u03bb (\u212b)")
        self.parent.bragg_edge_plot.setLabel("left", "Average Counts")

        o_get = Get(parent=self.parent, grand_parent=self.grand_parent)
        [linear_region_left_index, linear_region_right_index] = o_get.fitting_bragg_edge_linear_selection()

        # if self.grand_parent.fitting_bragg_edge_linear_selection == []:
        #     linear_region_left_index = 0
        #     linear_region_right_index = len(x_axis) - 1
        #     self.grand_parent.fitting_bragg_edge_linear_selection = [linear_region_left_index,
        #                                                              linear_region_right_index]
        #
        # else:
        #     [linear_region_left_index, linear_region_right_index] = \
        #         self.grand_parent.fitting_bragg_edge_linear_selection

        lr_left = x_axis[linear_region_left_index]
        lr_right = x_axis[linear_region_right_index]

        linear_region_range = [lr_left, lr_right]

        if self.parent.fitting_lr is None:
            lr = pg.LinearRegionItem(
                values=linear_region_range,
                orientation="vertical",
                brush=None,
                movable=True,
                bounds=None,
            )
            lr.setZValue(-10)
            lr.sigRegionChangeFinished.connect(self.parent.bragg_edge_linear_region_changed)
            lr.sigRegionChanged.connect(self.parent.bragg_edge_linear_region_changing)
            self.parent.bragg_edge_plot.addItem(lr)
            self.parent.fitting_lr = lr

        else:
            lr = self.parent.fitting_lr
            lr.setRegion(linear_region_range)
            self.parent.bragg_edge_plot.addItem(lr)

        display_fitting = True
        if display_fitting:
            parameters = self.get_average_parameters_activated()

            _advanced_fitting_mode = self.parent.ui.advanced_table_checkBox.isChecked()

            d_spacing = parameters["d_spacing"]
            alpha = parameters["alpha"]
            sigma = parameters["sigma"]
            a1 = parameters["a1"]
            a2 = parameters["a2"]
            if _advanced_fitting_mode:
                a5 = parameters["a5"]
                a6 = parameters["a6"]

            if np.isnan(d_spacing) or np.isnan(alpha) or np.isnan(sigma) or np.isnan(a1) or np.isnan(a2):
                return

            fit_x_axis = np.linspace(lr_left, lr_right, num=100)
            if _advanced_fitting_mode:
                fit_y_axis = [advanced_fit(x, d_spacing, alpha, sigma, a1, a2, a5, a6) for x in fit_x_axis]
            else:
                fit_y_axis = [basic_fit(x, d_spacing, alpha, sigma, a1, a2) for x in fit_x_axis]

            # fit_y_axis *= nbr_index_selected #FIXME

            self.parent.bragg_edge_plot.plot(fit_x_axis, fit_y_axis, pen="r")

    def get_average_parameters_activated(self):
        table_dictionary = self.grand_parent.march_table_dictionary

        d_spacing = []
        alpha = []
        sigma = []
        a1 = []
        a2 = []
        a5 = []
        a6 = []

        for _index in table_dictionary.keys():
            _entry = table_dictionary[_index]

            if _entry["active"]:
                _d_spacing = _entry["d_spacing"]["val"]
                _alpha = _entry["alpha"]["val"]
                _sigma = _entry["sigma"]["val"]
                _a1 = _entry["a1"]["val"]
                _a2 = _entry["a2"]["val"]
                _a5 = _entry["a5"]["val"]
                _a6 = _entry["a6"]["val"]

                d_spacing.append(_d_spacing)
                alpha.append(_alpha)
                sigma.append(_sigma)
                a1.append(_a1)
                a2.append(_a2)
                a5.append(_a5)
                a6.append(_a6)

        mean_d_spacing = self.get_mean_value(d_spacing)
        mean_alpha = self.get_mean_value(alpha)
        mean_sigma = self.get_mean_value(sigma)
        mean_a1 = self.get_mean_value(a1)
        mean_a2 = self.get_mean_value(a2)
        mean_a5 = self.get_mean_value(a5)
        mean_a6 = self.get_mean_value(a6)

        return {
            "d_spacing": mean_d_spacing,
            "alpha": mean_alpha,
            "sigma": mean_sigma,
            "a1": mean_a1,
            "a2": mean_a2,
            "a5": mean_a5,
            "a6": mean_a6,
        }

    def get_mean_value(self, array=None):
        if len(array) == 0:
            return np.nan
        else:
            return np.mean(array)



#!/usr/bin/env python
"""
EventHandler class for handling the events in the fitting tab.
"""

from qtpy import QtCore
from qtpy.QtWidgets import QApplication, QTableWidgetSelectionRange

from ibeatles.fitting.filling_table_handler import FillingTableHandler
from ibeatles.fitting.selected_bin_handler import SelectedBinsHandler
from ibeatles.table_dictionary.table_dictionary_handler import (
    TableDictionaryHandler,
)
from ibeatles.utilities.table_handler import TableHandler


class EventHandler:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def column_value_table_clicked(self, column):
        """
        to make sure that if the val or err column is selected, or unselected, the other
        column behave the same
        """
        if column < 5:
            return

        _item0 = self.grand_parent.fitting_ui.ui.value_table.item(0, column)
        state_column_clicked = self.grand_parent.fitting_ui.ui.value_table.isItemSelected(_item0)

        if column % 2 == 0:
            col1 = column - 1
            col2 = column
        else:
            col1 = column
            col2 = column + 1

        nbr_row = self.grand_parent.fitting_ui.ui.value_table.rowCount()
        range_selected = QTableWidgetSelectionRange(0, col1, nbr_row - 1, col2)
        self.grand_parent.fitting_ui.ui.value_table.setRangeSelected(range_selected, state_column_clicked)

    def column_header_table_clicked(self, column):
        _value_table_column = self.parent.header_value_tables_match.get(column, -1)
        nbr_row = self.grand_parent.fitting_ui.ui.value_table.rowCount()

        # if both col already selected, unselect them
        col_already_selected = False
        _item1 = self.grand_parent.fitting_ui.ui.value_table.item(0, _value_table_column[0])
        _item2 = self.grand_parent.fitting_ui.ui.value_table.item(0, _value_table_column[-1])

        if _item1.isSelected() and _item2.isSelected():
            col_already_selected = True

        if column in [2, 3]:
            selection = self.grand_parent.fitting_ui.ui.value_table.selectedRanges()
            col_already_selected = False
            for _select in selection:
                if column in [_select.leftColumn(), _select.rightColumn()]:
                    col_already_selected = True
                    break

        from_col = _value_table_column[0]
        to_col = _value_table_column[-1]

        range_selected = QTableWidgetSelectionRange(0, from_col, nbr_row - 1, to_col)
        self.grand_parent.fitting_ui.ui.value_table.setRangeSelected(range_selected, not col_already_selected)

    def resizing_header_table(self, index_column, new_size):
        if index_column < 5:
            self.parent.ui.value_table.setColumnWidth(index_column, new_size)
        else:
            new_half_size = int(new_size / 2)
            index1 = (index_column - 5) * 2 + 5
            index2 = index1 + 1
            self.parent.ui.value_table.setColumnWidth(index1, new_half_size)
            self.parent.ui.value_table.setColumnWidth(index2, new_half_size)

    def resizing_value_table(self, index_column, new_size):
        if index_column < 5:
            self.parent.ui.header_table.setColumnWidth(index_column, new_size)
        else:
            if (index_column % 2) == 1:
                right_new_size = self.parent.ui.value_table.columnWidth(index_column + 1)
                index_header = int(index_column - 5) / 2 + 5
                self.parent.ui.header_table.setColumnWidth(index_header, new_size + right_new_size)

            else:
                left_new_size = self.parent.ui.value_table.columnWidth(index_column - 1)
                index_header = int(index_column - 6) / 2 + 5
                self.parent.ui.header_table.setColumnWidth(index_header, new_size + left_new_size)

    def check_state_of_step3_button(self):
        """The step1 button should be enabled if at least one row of the big table
        is activated and display in the 1D plot"""
        o_table = TableDictionaryHandler(parent=self.parent, grand_parent=self.grand_parent)
        is_at_least_one_row_activated = o_table.is_at_least_one_row_activated()
        self.parent.ui.step3_button.setEnabled(is_at_least_one_row_activated)
        self.parent.ui.step2_instruction_label.setEnabled(is_at_least_one_row_activated)

    def check_state_of_step4_button(self):
        self.parent.ui.step4_button.setEnabled(self.parent.is_ready_to_fit)

    def active_button_state_changed(self, row_clicked):
        """
        status: 0: off
                2: on
        """
        QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        update_lock_flag = False
        if self.grand_parent.advanced_selection_ui:
            self.grand_parent.advanced_selection_ui.ui.selection_table.blockSignals(True)

        self.parent.mirror_state_of_widgets(column=3, row_clicked=row_clicked)
        self.check_state_of_step3_button()

        o_bin_handler = SelectedBinsHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_bin_handler.update_bins_selected()
        self.parent.update_bragg_edge_plot()
        o_bin_handler.update_bins_locked()

        if self.grand_parent.advanced_selection_ui:
            self.grand_parent.advanced_selection_ui.update_selection_table()
            if update_lock_flag:
                self.grand_parent.advanced_selection_ui.update_lock_table()
            self.grand_parent.advanced_selection_ui.ui.selection_table.blockSignals(False)

        QApplication.restoreOverrideCursor()

    def mirror_state_of_widgets(self, column=2, row_clicked=0):
        # perform same status on all rows and save it in table_dictionary
        label_column = "active" if column == 3 else "lock"

        o_table = TableHandler(table_ui=self.parent.ui.value_table)
        o_table.add_this_row_to_selection(row=row_clicked)
        list_row_selected = o_table.get_rows_of_table_selected()

        o_table_handler = TableDictionaryHandler(grand_parent=self.grand_parent, parent=self.parent)
        is_this_row_checked = o_table_handler.is_this_row_checked(row=row_clicked, column=column)

        for _row in list_row_selected:
            self.grand_parent.march_table_dictionary[str(_row)][label_column] = is_this_row_checked
            if _row == row_clicked:
                continue
            _widget = o_table.get_widget(row=_row, column=column)
            _widget.blockSignals(True)
            _widget.setChecked(is_this_row_checked)
            _widget.blockSignals(False)

    def lock_button_state_changed(self, status, row_clicked):
        """
        All the row selected should mirror the state of this button

        status: 0: off
                2: on
        """
        update_selection_flag = False

        if self.grand_parent.advanced_selection_ui:
            self.grand_parent.advanced_selection_ui.ui.lock_table.blockSignals(True)

        if status == 0:
            status = False
        else:
            status = True

        self.mirror_state_of_widgets(column=2, row_clicked=row_clicked)

        # hide this row if status is False and user only wants to see locked items
        o_filling_handler = FillingTableHandler(grand_parent=self.grand_parent, parent=self.parent)
        if (status is False) and (o_filling_handler.get_row_to_show_state() == "lock"):
            self.grand_parent.fitting_ui.ui.value_table.hideRow(row_clicked)

        o_bin_handler = SelectedBinsHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_bin_handler.update_bins_locked()
        self.parent.update_bragg_edge_plot()
        o_bin_handler.update_bins_selected()

        if self.grand_parent.advanced_selection_ui:
            self.grand_parent.advanced_selection_ui.update_lock_table()
            if update_selection_flag:
                self.grand_parent.advanced_selection_ui.update_selection_table()
            self.grand_parent.advanced_selection_ui.ui.lock_table.blockSignals(False)

    def update_image_view_selection(self):
        o_bin_handler = SelectedBinsHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_bin_handler.update_bins_selected()

    def update_image_view_lock(self):
        o_bin_handler = SelectedBinsHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_bin_handler.update_bins_locked()



#!/usr/bin/env python
"""
Manual event handler
"""

import logging

import numpy as np
import pyqtgraph as pg

from ibeatles.tools.tof_bin import TO_ANGSTROMS_UNITS, TO_MICROS_UNITS
from ibeatles.tools.tof_bin.plot import Plot
from ibeatles.tools.tof_bin.utilities.get import Get
from ibeatles.tools.utilities import TimeSpectraKeys
from ibeatles.utilities.math_tools import get_index_of_closest_match
from ibeatles.utilities.string import format_str
from ibeatles.utilities.table_handler import TableHandler

FILE_INDEX_BIN_MARGIN = 0.5
UNSELECTED_BIN = (0, 0, 200, 50)
SELECTED_BIN = (0, 200, 0, 50)


class ManualEventHandler:
    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger("maverick")

        self.tof_bin_margin = (
            self.parent.time_spectra[TimeSpectraKeys.tof_array][1]
            - self.parent.time_spectra[TimeSpectraKeys.tof_array][0]
        ) / 2.0

        self.lambda_bin_margin = (
            self.parent.time_spectra[TimeSpectraKeys.lambda_array][1]
            - self.parent.time_spectra[TimeSpectraKeys.lambda_array][0]
        ) / 2

    def refresh_manual_tab(self):
        """refresh the right plot with profile + bin selected when the manual tab is selected"""
        o_plot = Plot(parent=self.parent)
        o_plot.refresh_profile_plot_and_clear_bins()

    # def save_bins_in_all_units(self):
    #
    #     bins = self.parent.manual_bins[time_spectra_x_axis_name]
    #     if not bins:
    #         return
    #
    #     dict_of_bins_item = {}
    #     for _index, _bin in enumerate(bins):
    #         if len(_bin) == 0:
    #             continue
    #
    #         if time_spectra_x_axis_name == TimeSpectraKeys.file_index_array:
    #             scale_bin = [_bin[0] - FILE_INDEX_BIN_MARGIN, _bin[-1] + FILE_INDEX_BIN_MARGIN]
    #
    #         elif time_spectra_x_axis_name == TimeSpectraKeys.tof_array:
    #             scale_bin = [_bin[0] - self.tof_bin_margin, _bin[-1] + self.tof_bin_margin]
    #             scale_bin = [_value * TO_MICROS_UNITS for _value in scale_bin]
    #
    #         else:
    #             scale_bin = [
    #                 _bin[0] - self.lambda_bin_margin,
    #                 _bin[-1] + self.lambda_bin_margin,
    #             ]
    #             scale_bin = [_value * TO_ANGSTROMS_UNITS for _value in scale_bin]
    #
    #         item = pg.LinearRegionItem(
    #             values=scale_bin,
    #             orientation="vertical",
    #             brush=UNSELECTED_BIN,
    #             movable=True,
    #             bounds=None,
    #         )
    #         item.setZValue(-10)
    #         self.parent.bin_profile_view.addItem(item)
    #         item.sigRegionChangeFinished.connect(self.parent.bin_manual_region_changed)
    #         item.sigRegionChanged.connect(self.parent.bin_manual_region_changing)
    #         dict_of_bins_item[_index] = item
    #
    #     self.parent.dict_of_bins_item = dict_of_bins_item

    def add_bin(self):
        o_get = Get(parent=self.parent)
        time_spectra_x_axis_name = o_get.x_axis_selected()
        x_axis = self.parent.time_spectra[time_spectra_x_axis_name]

        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        last_row = o_table.row_count()

        if time_spectra_x_axis_name == TimeSpectraKeys.file_index_array:
            default_bin = [
                x_axis[0] - FILE_INDEX_BIN_MARGIN,
                x_axis[0] + FILE_INDEX_BIN_MARGIN,
            ]
        elif time_spectra_x_axis_name == TimeSpectraKeys.tof_array:
            default_bin = [
                x_axis[0] - self.tof_bin_margin,
                x_axis[0] + self.tof_bin_margin,
            ]
            default_bin = [_value * TO_MICROS_UNITS for _value in default_bin]
        else:
            default_bin = [
                x_axis[0] - self.lambda_bin_margin,
                x_axis[0] + self.lambda_bin_margin,
            ]
            default_bin = [_value * TO_ANGSTROMS_UNITS for _value in default_bin]

        manual_bins = self.parent.manual_bins
        if manual_bins[TimeSpectraKeys.file_index_array] is None:
            manual_bins[TimeSpectraKeys.file_index_array] = [
                self.parent.time_spectra[TimeSpectraKeys.file_index_array][0]
            ]
            manual_bins[TimeSpectraKeys.tof_array] = [[self.parent.time_spectra[TimeSpectraKeys.tof_array][0]]]
            manual_bins[TimeSpectraKeys.lambda_array] = [[self.parent.time_spectra[TimeSpectraKeys.lambda_array][0]]]
        else:
            manual_bins[TimeSpectraKeys.file_index_array].append(
                [self.parent.time_spectra[TimeSpectraKeys.file_index_array][0]]
            )
            manual_bins[TimeSpectraKeys.tof_array].append([self.parent.time_spectra[TimeSpectraKeys.tof_array][0]])
            manual_bins[TimeSpectraKeys.lambda_array].append(
                [self.parent.time_spectra[TimeSpectraKeys.lambda_array][0]]
            )
        self.parent.manual_bins = manual_bins

        item = pg.LinearRegionItem(
            values=default_bin,
            orientation="vertical",
            brush=SELECTED_BIN,
            movable=True,
            bounds=None,
        )
        item.setZValue(-10)
        item.sigRegionChangeFinished.connect(self.parent.bin_manual_region_changed)
        item.sigRegionChanged.connect(self.parent.bin_manual_region_changing)

        self.parent.bin_profile_view.addItem(item)
        # dict_of_bins_item[last_row] = item
        # self.parent.dict_of_bins_item = dict_of_bins_item
        self.parent.list_of_manual_bins_item.append(item)

        # add new entry in table
        o_table.insert_empty_row(last_row)

        o_table.insert_item(row=last_row, column=0, value=f"{last_row}", editable=False)

        _file_index = self.parent.time_spectra[TimeSpectraKeys.file_index_array][0]
        o_table.insert_item(row=last_row, column=1, value=_file_index, editable=False)

        _tof = self.parent.time_spectra[TimeSpectraKeys.tof_array][0] * TO_MICROS_UNITS
        o_table.insert_item(row=last_row, column=2, value=_tof, format_str="{:.2f}", editable=False)

        _lambda = self.parent.time_spectra[TimeSpectraKeys.lambda_array][0] * TO_ANGSTROMS_UNITS
        o_table.insert_item(row=last_row, column=3, value=_lambda, format_str="{:.3f}", editable=False)

    def clear_all_items(self):
        list_of_manually_bins_item = self.parent.list_of_manual_bins_item
        for _item in list_of_manually_bins_item:
            self.parent.bin_profile_view.removeItem(_item)

        self.parent.list_of_manual_bins_item = None

    def display_all_items(self):
        list_of_manually_bins_item = self.parent.list_of_manual_bins_item
        for _item in list_of_manually_bins_item:
            self.parent.bin_profile_view.addItem(_item)

    def populate_table_with_this_table(self, table=None):
        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        o_table.remove_all_rows()

        file_index_array = table[TimeSpectraKeys.file_index_array]
        tof_array = table[TimeSpectraKeys.tof_array]
        lambda_array = table[TimeSpectraKeys.lambda_array]

        self.parent.list_of_manual_bins_item = []

        if file_index_array is None:
            return

        _row = 0
        for _index, _bin in enumerate(file_index_array):
            if not _bin:
                continue

            o_table.insert_empty_row(_row)

            o_table.insert_item(row=_row, column=0, value=f"{_row}", editable=False)

            _file_index = _bin
            _file_index_formatted = format_str(
                _file_index,
                format_str="{:d}",
                factor=1,
                data_type=TimeSpectraKeys.file_index_array,
            )
            o_table.insert_item(row=_row, column=1, value=_file_index_formatted, editable=False)

            _tof = tof_array[_index]
            _tof_formatted = format_str(
                _tof,
                format_str="{:.2f}",
                factor=TO_MICROS_UNITS,
                data_type=TimeSpectraKeys.tof_array,
            )
            o_table.insert_item(row=_row, column=2, value=_tof_formatted, editable=False)

            _lambda = lambda_array[_index]
            _lambda_formatted = format_str(
                _lambda,
                format_str="{:.3f}",
                factor=TO_ANGSTROMS_UNITS,
                data_type=TimeSpectraKeys.lambda_array,
            )
            o_table.insert_item(row=_row, column=3, value=_lambda_formatted, editable=False)

            item = self.add_bin_in_plot(row=_row, file_index_bin=_bin, tof_bin=_tof, lambda_bin=_lambda)

            self.parent.list_of_manual_bins_item.append(item)

            _row += 1

        o_table.select_rows([0])

    def populate_table_with_auto_mode(self):
        o_get = Get(parent=self.parent)
        bins = o_get.auto_bins_currently_activated()
        self.parent.manual_bins = bins
        self.populate_table_with_this_table(table=bins)

    def add_bin_in_plot(self, row=0, file_index_bin=None, tof_bin=None, lambda_bin=None):
        o_get = Get(parent=self.parent)
        current_x_axis = o_get.x_axis_selected()
        if current_x_axis == TimeSpectraKeys.file_index_array:
            bin = file_index_bin
            bin_size = [bin[0] - FILE_INDEX_BIN_MARGIN, bin[-1] + FILE_INDEX_BIN_MARGIN]
        elif current_x_axis == TimeSpectraKeys.tof_array:
            bin = tof_bin
            bin_size = [bin[0] - self.tof_bin_margin, bin[-1] + self.tof_bin_margin]
        elif current_x_axis == TimeSpectraKeys.lambda_array:
            bin = lambda_bin
            bin_size = [
                bin[0] - self.lambda_bin_margin,
                bin[1] + self.lambda_bin_margin,
            ]
        else:
            raise NotImplementedError("x_axis not implemented!")

        if row == 0:
            brush_selection = SELECTED_BIN
        else:
            brush_selection = UNSELECTED_BIN

        item = pg.LinearRegionItem(
            values=bin_size,
            orientation="vertical",
            brush=brush_selection,
            movable=True,
            bounds=None,
        )
        item.setZValue(-10)
        item.sigRegionChangeFinished.connect(self.parent.bin_manual_region_changed)
        self.parent.bin_profile_view.addItem(item)

        return item

    def bin_manually_moved(self, item_id=None):
        self.bin_manually_moving(item_id=item_id)

        # 1. using region selected threshold, and the current axis, find the snapping left and right indexes
        #    and save them into a manual_snapping_indexes_bins = {0: [0, 3], 1: [1, 10], ..}
        self.record_snapping_indexes_bin()

        # 2. using those indexes create the ranges for each bins and for each time axis and save those in
        #    self.parent.manual_bins['file_index_array': [[0, 1, 2, 3], [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], ...], ...]
        self.create_all_ranges()

        # 3. update table
        self.update_table()

    def bin_manually_moving(self, item_id=None):
        o_get = Get(parent=self.parent)
        working_row = o_get.manual_working_row(working_item_id=item_id)
        self.select_working_row(working_row=working_row)

    def select_working_row(self, working_row=0):
        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        o_table.block_signals(True)
        o_table.select_rows(list_of_rows=[working_row])
        o_table.block_signals(False)

    def update_table(self):
        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        o_table.block_signals(True)

        file_index_array = self.parent.manual_bins[TimeSpectraKeys.file_index_array]
        tof_array = self.parent.manual_bins[TimeSpectraKeys.tof_array]
        lambda_array = self.parent.manual_bins[TimeSpectraKeys.lambda_array]

        for _row, list_runs in enumerate(file_index_array):
            list_runs_formatted = format_str(
                list_runs,
                format_str="{:d}",
                factor=1,
                data_type=TimeSpectraKeys.file_index_array,
            )
            o_table.set_item_with_str(row=_row, column=1, cell_str=list_runs_formatted)

            list_tof = tof_array[_row]
            list_tof_formatted = format_str(
                list_tof,
                format_str="{:.2f}",
                factor=TO_MICROS_UNITS,
                data_type=TimeSpectraKeys.tof_array,
            )
            o_table.set_item_with_str(row=_row, column=2, cell_str=list_tof_formatted)

            list_lambda = lambda_array[_row]
            list_lambda_formatted = format_str(
                list_lambda,
                format_str="{:.3f}",
                factor=TO_ANGSTROMS_UNITS,
                data_type=TimeSpectraKeys.lambda_array,
            )
            o_table.set_item_with_str(row=_row, column=3, cell_str=list_lambda_formatted)

        o_table.block_signals(False)

    # def create_all_ranges(self):
    #     manual_snapping_indexes_bins = self.parent.manual_snapping_indexes_bins
    #
    #     file_index_array = {}
    #     tof_array = {}
    #     lambda_array = {}
    #
    #     for _bin in manual_snapping_indexes_bins.keys():
    #         left_index, right_index = manual_snapping_indexes_bins[_bin]
    #
    #         # tof_array
    #         bins_file_index_array = self.parent.time_spectra[TimeSpectraKeys.file_index_array]
    #         bins_file_index_range = bins_file_index_array[left_index: right_index + 1]
    #         file_index_array[_bin] = bins_file_index_range
    #
    #         # tof_array
    #         bins_tof_array = self.parent.time_spectra[TimeSpectraKeys.tof_array]
    #         bins_tof_range = bins_tof_array[left_index: right_index + 1]
    #         tof_array[_bin] = bins_tof_range
    #
    #         # lambda_array
    #         bins_lambda_array = self.parent.time_spectra[TimeSpectraKeys.lambda_array]
    #         bins_lambda_range = bins_lambda_array[left_index: right_index + 1]
    #         lambda_array[_bin] = bins_lambda_range
    #
    #     self.parent.manual_bins[TimeSpectraKeys.file_index_array] = file_index_array
    #     self.parent.manual_bins[TimeSpectraKeys.tof_array] = tof_array
    #     self.parent.manual_bins[TimeSpectraKeys.lambda_array] = lambda_array

    def create_all_ranges(self):
        manual_snapping_indexes_bins = self.parent.manual_snapping_indexes_bins

        file_index_array = []
        tof_array = []
        lambda_array = []

        for _bin in manual_snapping_indexes_bins.keys():
            left_index, right_index = manual_snapping_indexes_bins[_bin]

            # tof_array
            bins_file_index_array = list(self.parent.time_spectra[TimeSpectraKeys.file_index_array])
            bins_file_index_range = bins_file_index_array[left_index : right_index + 1]
            file_index_array.append(bins_file_index_range)

            # tof_array
            bins_tof_array = self.parent.time_spectra[TimeSpectraKeys.tof_array]
            bins_tof_range = bins_tof_array[left_index : right_index + 1]
            tof_array.append(bins_tof_range)

            # lambda_array
            bins_lambda_array = self.parent.time_spectra[TimeSpectraKeys.lambda_array]
            bins_lambda_range = bins_lambda_array[left_index : right_index + 1]
            lambda_array.append(bins_lambda_range)

        self.parent.manual_bins[TimeSpectraKeys.file_index_array] = file_index_array
        self.parent.manual_bins[TimeSpectraKeys.tof_array] = tof_array
        self.parent.manual_bins[TimeSpectraKeys.lambda_array] = lambda_array

    # def update_manual_snapping_indexes_bins(self):
    #     """this will take the manual_bins list, use the x-axis currently selected and replace the
    #     manual_snapping_indexes_bins [[left, right], [left, right]....] of the current xaxis"""
    #     manual_snapping_indexes_bin = {}
    #     for _row, _bin_range in enumerate(self.parent.manual_bins[TimeSpectraKeys.file_index_array]):
    #         print(f"{_bin_range =}")
    #         manual_snapping_indexes_bin[_row] = [_bin_range[0], _bin_range[-1]]
    #     self.parent.manual_snapping_indexes_bins = manual_snapping_indexes_bin
    #     print(f"{self.parent.manual_snapping_indexes_bins =}")

    def update_items_displayed(self):
        """
        this will remove the old item and put the new one with the edges snap to the x-axis
        """
        o_get = Get(parent=self.parent)
        x_axis_type_selected = o_get.x_axis_selected()
        if x_axis_type_selected == TimeSpectraKeys.tof_array:
            factor = TO_MICROS_UNITS
        elif x_axis_type_selected == TimeSpectraKeys.lambda_array:
            factor = TO_ANGSTROMS_UNITS
        else:
            factor = 1

        x_axis = self.parent.manual_bins[x_axis_type_selected]
        if x_axis is None:
            return

        if len(self.parent.list_of_manual_bins_item) == 0:
            return

        list_of_manual_bins_item = []
        for _row, _x in enumerate(x_axis):
            left_value_checked = _x[0] * factor
            right_value_checked = _x[-1] * factor

            _item = self.parent.list_of_manual_bins_item[_row]

            self.parent.bin_profile_view.removeItem(_item)

            item = pg.LinearRegionItem(
                values=[left_value_checked, right_value_checked],
                orientation="vertical",
                brush=SELECTED_BIN,
                movable=True,
                bounds=None,
            )
            item.setZValue(-10)
            item.sigRegionChangeFinished.connect(self.parent.bin_manual_region_changed)
            item.sigRegionChanged.connect(self.parent.bin_manual_region_changing)
            self.parent.bin_profile_view.addItem(item)
            list_of_manual_bins_item.append(item)

        self.parent.list_of_manual_bins_item = list_of_manual_bins_item

    def record_snapping_indexes_bin(self):
        """
        This will check each bin from the manual table and move, if necessary, any of the edges
        to snap to the closet x-axis values
        """
        manual_snapping_indexes_bins = {}
        for _row, _item in enumerate(self.parent.list_of_manual_bins_item):
            [left, right] = _item.getRegion()

            # bring left and right to closest correct values
            left_value_checked, right_value_checked = self.checked_range(left=left, right=right)
            manual_snapping_indexes_bins[_row] = [
                left_value_checked,
                right_value_checked,
            ]

        self.parent.manual_snapping_indexes_bins = manual_snapping_indexes_bins

    def margin(self, axis_type=TimeSpectraKeys.file_index_array):
        if axis_type == TimeSpectraKeys.file_index_array:
            return FILE_INDEX_BIN_MARGIN
        elif axis_type == TimeSpectraKeys.tof_array:
            return self.tof_bin_margin
        elif axis_type == TimeSpectraKeys.lambda_array:
            return self.lambda_bin_margin
        else:
            raise NotImplementedError(f"axis type {axis_type} not implemented!")

    def checked_range(self, left=0, right=0):
        """this method makes sure that the left and right values stay within the maximum range of the data
        for the current axis selected"""
        o_get = Get(parent=self.parent)
        x_axis_type_selected = o_get.x_axis_selected()
        x_axis = self.parent.time_spectra[x_axis_type_selected]

        if x_axis_type_selected == TimeSpectraKeys.tof_array:
            factor = TO_MICROS_UNITS
        elif x_axis_type_selected == TimeSpectraKeys.lambda_array:
            factor = TO_ANGSTROMS_UNITS
        else:
            factor = 1

        left /= factor
        right /= factor

        if left < x_axis[0]:
            left = x_axis[0]

        if right >= x_axis[-1]:
            right = x_axis[-1]

        index_clean_left_value = get_index_of_closest_match(array_to_look_for=x_axis, value=left, left_margin=True)
        index_clean_right_value = get_index_of_closest_match(array_to_look_for=x_axis, value=right, left_margin=False)

        return (
            np.min([index_clean_left_value, index_clean_right_value]),
            np.max([index_clean_left_value, index_clean_right_value]),
        )



#!/usr/bin/env python
"""
GUI Handler (step 2)
"""

import numpy as np
import pyqtgraph as pg

from ibeatles.step2.get import Get
from ibeatles.step2.plot import Step2Plot


class CustomAxis(pg.AxisItem):
    def tickStrings(self, values, scale, spacing):
        values[values == 0] = np.nan  # remove 0 before division
        return ["{:.4f}".format(1.0 / i) for i in values]


class Step2GuiHandler(object):
    col_width = [70, 50, 50, 50, 50]

    def __init__(self, parent=None):
        self.parent = parent

    def update_widgets(self):
        o_step2_plot = Step2Plot(parent=self.parent)
        o_step2_plot.prepare_data()
        o_step2_plot.display_image()
        o_step2_plot.init_roi_table()
        self.check_run_normalization_button()

    def check_add_remove_roi_buttons(self):
        nbr_row = self.parent.ui.normalization_tableWidget.rowCount()
        if nbr_row == 0:
            _status_remove = False
        else:
            _status_remove = True

        self.parent.ui.normalization_remove_roi_button.setEnabled(_status_remove)

    def check_run_normalization_button(self):
        o_get = Get(parent=self.parent)
        _status = o_get.status_of_run_normalization_button()
        self.parent.ui.normalization_button.setEnabled(_status)
        self.update_instructions()

    def enable_xaxis_button(self, tof_flag=True):
        list_ui = [
            self.parent.step2_ui["xaxis_file_index"],
            self.parent.step2_ui["xaxis_lambda"],
            self.parent.step2_ui["xaxis_tof"],
        ]

        if tof_flag:
            for _ui in list_ui:
                _ui.setEnabled(True)
        else:
            list_ui[1].setEnabled(False)
            list_ui[2].setEnabled(False)
            list_ui[0].setChecked(True)

    def update_instructions(self):
        o_get = Get(parent=self.parent)
        _status = o_get.status_of_run_normalization_button()
        if not _status:
            instructions = (
                '<font color="red">Activate</font> at least one <font color="red">Background ROI</font> '
                "to enable the Normalization switch!</html>"
            )
        else:
            instructions = 'A <font color="red">Background ROI</font> will improve the normalization!</html>'
        self.parent.ui.normalization_instructions.setHtml(instructions)



import numpy as np
from lmfit import Model
from qtpy import QtGui
from qtpy.QtWidgets import QApplication, QTableWidgetItem

from ..fitting.fitting_functions import advanced_fit, basic_fit


class ResultValueError(object):
    def __init__(self, result=None):
        self.result = result

    def get_value_err(self, tag=""):
        value = self.result.params[tag].value
        error = self.result.params[tag].stderr

        return [value, error]


class FittingJobHandler(object):
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def run_story(self):
        table_fitting_story_dictionary = self.grand_parent.table_fitting_story_dictionary
        table_dictionary = self.grand_parent.march_table_dictionary
        nbr_entry = len(table_fitting_story_dictionary)

        _advanced_fitting_mode = self.grand_parent.fitting_ui.ui.advanced_table_checkBox.isChecked()

        # define fitting equation
        if _advanced_fitting_mode:
            gmodel = Model(advanced_fit, missing="drop")  # do not considerate the np.nan
        else:
            gmodel = Model(basic_fit, missing="drop")

        # index of selection in bragg edge plot
        [left_index, right_index] = self.grand_parent.fitting_bragg_edge_linear_selection

        # retrieve image
        data_2d = np.array(self.grand_parent.data_metadata["normalized"]["data"])
        full_x_axis = self.parent.bragg_edge_data["x_axis"]
        x_axis = np.array(full_x_axis[left_index:right_index], dtype=float)

        self.grand_parent.fitting_story_ui.eventProgress.setValue(0)
        self.grand_parent.fitting_story_ui.eventProgress.setMaximum(nbr_entry)
        self.grand_parent.fitting_story_ui.eventProgress.setVisible(True)
        self.grand_parent.fitting_story_ui.eventProgress2.setVisible(True)

        for _entry_index in table_fitting_story_dictionary.keys():
            self.status_of_row(row=_entry_index, status="IN PROGRESS")

            _entry = table_fitting_story_dictionary[_entry_index]
            d_spacing_flag = _entry["d_spacing"]
            alpha_flag = _entry["alpha"]
            sigma_flag = _entry["sigma"]
            a1_flag = _entry["a1"]
            a2_flag = _entry["a2"]

            if _advanced_fitting_mode:
                a5_flag = _entry["a5"]
                a6_flag = _entry["a6"]

            self.grand_parent.fitting_story_ui.eventProgress2.setValue(0)
            self.grand_parent.fitting_story_ui.eventProgress2.setMaximum(len(table_dictionary))

            # loop over all the bins
            progress_bar_index = 0
            for _bin_index in table_dictionary:
                _bin_entry = table_dictionary[_bin_index]

                if _bin_entry["active"]:
                    # define status of variables
                    params = gmodel.make_params()

                    _d_spacing = _bin_entry["d_spacing"]["val"]
                    params.add("d_spacing", value=_d_spacing, vary=d_spacing_flag)

                    _sigma = _bin_entry["sigma"]["val"]
                    params.add("sigma", value=_sigma, vary=sigma_flag)

                    _alpha = _bin_entry["alpha"]["val"]
                    params.add("alpha", value=_alpha, vary=alpha_flag)

                    _a1 = _bin_entry["a1"]["val"]
                    params.add("a1", value=_a1, vary=a1_flag)

                    _a2 = _bin_entry["a2"]["val"]
                    params.add("a2", value=_a2, vary=a2_flag)

                    if _advanced_fitting_mode:
                        _a5 = _bin_entry["a5"]["val"]
                        params.add("a5", value=_a5, vary=a5_flag)

                        _a6 = _bin_entry["a6"]["val"]
                        params.add("a6", value=_a6, vary=a6_flag)

                    _bin_x0 = _bin_entry["bin_coordinates"]["x0"]
                    _bin_x1 = _bin_entry["bin_coordinates"]["x1"]
                    _bin_y0 = _bin_entry["bin_coordinates"]["y0"]
                    _bin_y1 = _bin_entry["bin_coordinates"]["y1"]

                    y_axis = data_2d[
                        left_index:right_index,
                        _bin_x0:_bin_x1,
                        _bin_y0:_bin_y1,
                    ]  # noqa: E124

                    # y_axis = y_axis.sum(axis=1)
                    # y_axis = np.array(y_axis.sum(axis=1), dtype=float)
                    y_axis = np.nanmean(y_axis, axis=1)
                    y_axis = np.array(np.nanmean(y_axis, axis=1), dtype=float)

                    try:
                        result = gmodel.fit(y_axis, params, t=x_axis)
                    except ValueError:
                        self.status_of_row(row=_entry_index, status="FAILED")
                        # FIXME
                        # show dialog message that informs that fitting did not converge
                        # tell which step failed

                    _o_result = ResultValueError(result=result)
                    if d_spacing_flag:
                        [value, error] = _o_result.get_value_err(tag="d_spacing")
                        _bin_entry["d_spacing"]["val"] = value
                        _bin_entry["d_spacing"]["err"] = error

                    if sigma_flag:
                        [value, error] = _o_result.get_value_err(tag="sigma")
                        _bin_entry["sigma"]["val"] = value
                        _bin_entry["sigma"]["err"] = error

                    if alpha_flag:
                        tag = "alpha"
                        [value, error] = _o_result.get_value_err(tag=tag)
                        _bin_entry[tag]["val"] = value
                        _bin_entry[tag]["err"] = error

                    if a1_flag:
                        tag = "a1"
                        [value, error] = _o_result.get_value_err(tag=tag)
                        _bin_entry[tag]["val"] = value
                        _bin_entry[tag]["err"] = error

                    if a2_flag:
                        tag = "a2"
                        [value, error] = _o_result.get_value_err(tag=tag)
                        _bin_entry[tag]["val"] = value
                        _bin_entry[tag]["err"] = error

                    if _advanced_fitting_mode:
                        if a5_flag:
                            tag = "a5"
                            [value, error] = _o_result.get_value_err(tag=tag)
                            _bin_entry[tag]["val"] = value
                            _bin_entry[tag]["err"] = error

                        if a6_flag:
                            tag = "a6"
                            [value, error] = _o_result.get_value_err(tag=tag)
                            _bin_entry[tag]["val"] = value
                            _bin_entry[tag]["err"] = error

                    table_dictionary[_bin_index] = _bin_entry

                progress_bar_index += 1
                self.grand_parent.fitting_story_ui.eventProgress2.setValue(progress_bar_index)
                QApplication.processEvents()

            self.status_of_row(row=_entry_index, status="DONE")

            self.grand_parent.fitting_story_ui.eventProgress.setValue(_entry_index + 1)
            QApplication.processEvents()

            self.grand_parent.march_table_dictionary = table_dictionary
            self.grand_parent.fitting_ui.re_fill_table()
            self.grand_parent.fitting_ui.update_bragg_edge_plot(update_selection=False)

        self.grand_parent.fitting_story_ui.eventProgress.setVisible(False)
        self.grand_parent.fitting_story_ui.eventProgress2.setVisible(False)

    def status_of_row(self, row=0, status="IN PROGRESS"):
        if status == "IN PROGRESS":
            _color = QtGui.QColor(0, 0, 255)  # blue
        elif status == "DONE":
            _color = QtGui.QColor(21, 190, 21)  # green
        elif status == "FAILED":
            status = "SOME ERRORS!"
            _color = QtGui.QColor(255, 0, 0)  # red
        else:
            _color = QtGui.QColor(0, 0, 0)  # black
        _item = QTableWidgetItem(status)
        # _item.setTextColor(_color)
        _brush = QtGui.QBrush(_color)
        _item.setForeground(_brush)

        self.grand_parent.fitting_story_ui.ui.story_table.setItem(row, 7, _item)
        QApplication.processEvents()



#!/usr/bin/env python
"""
General class for general settings from session file.
"""

from ibeatles.session import SessionSubKeys


class General:
    def __init__(self, parent=None):
        self.parent = parent
        self.session_dict = parent.session_dict

    def settings(self):
        if self.session_dict.get("log buffer size", None):
            pass
        else:
            self.session_dict[SessionSubKeys.log_buffer_size] = self.parent.default_session_dict[
                SessionSubKeys.log_buffer_size
            ]



#!/usr/bin/env python
# ---------------------------------------------------------------------------------------------
# Copyright (c) 2011-2014, Ryan Galloway (ryan@rsgalloway.com)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of the software nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# ---------------------------------------------------------------------------------------------
# docs and latest version available for download at
#   http://rsgalloway.github.com/qrangeslider
# ---------------------------------------------------------------------------------------------
import sys

from qtpy import QtCore, QtGui
from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QApplication,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QSplitter,
    QWidget,
)

__author__ = "Ryan Galloway <ryan@rsgalloway.com>"
__version__ = "0.1.1+mm+qt5+v"


# ---------------------------------------------------------------------------------------------
# SUMMARY
# ---------------------------------------------------------------------------------------------
"""The QRangeSlider class implements a horizontal and vertical range slider widget.
"""

# ---------------------------------------------------------------------------------------------
# TODO
# ---------------------------------------------------------------------------------------------

"""
  - smoother mouse move event handler
  - support splits and joins
  - verticle sliders
  - ticks
"""

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s  # noqa: E731

__all__ = ["QRangeSlider"]

DEFAULT_CSS = """
QRangeSlider * {
    border: 0px;
    padding: 0px;
}
QRangeSlider #Head {
    background: #222;
}
QRangeSlider #Span {
    background: #393;
}
QRangeSlider #Span:active {
    background: #3d3;
}
QRangeSlider #Tail {
    background: #222;
}
QRangeSlider > QSplitter::handle {
    background: #393;
}
QRangeSlider > QSplitter::handle:vertical {
    height: 4px;
}
QRangeSlider > QSplitter::handle:pressed {
    background: #ddd;
}

"""


def scale(val, src, dst):
    """
    Scale the given value from the scale of src to the scale of dst.
    """
    return int(((val - src[0]) / float(src[1] - src[0])) * (dst[1] - dst[0]) + dst[0])


class RangeSliderForm:
    """default range slider form"""

    def setupUi(self, Form, splitterWidth=4, vertical=False):
        Form.setObjectName(_fromUtf8("QRangeSlider"))
        if vertical:
            Form.resize(30, 300)
        else:
            Form.resize(300, 30)
        Form.setStyleSheet(_fromUtf8(DEFAULT_CSS))
        self.gridLayout = QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName(_fromUtf8("gridLayout"))
        self._splitter = QSplitter(Form)
        self._splitter.setHandleWidth(splitterWidth)
        self._splitter.setMinimumSize(QtCore.QSize(0, 0))
        self._splitter.setMaximumSize(QtCore.QSize(16777215, 16777215))
        if vertical:
            self._splitter.setOrientation(QtCore.Qt.Vertical)
        else:
            self._splitter.setOrientation(QtCore.Qt.Horizontal)
        self._splitter.setObjectName(_fromUtf8("splitter"))
        self._head = QGroupBox(self._splitter)
        self._head.setTitle(_fromUtf8(""))
        self._head.setObjectName(_fromUtf8("Head"))
        self._handle = QGroupBox(self._splitter)
        self._handle.setTitle(_fromUtf8(""))
        self._handle.setObjectName(_fromUtf8("Span"))
        self._tail = QGroupBox(self._splitter)
        self._tail.setTitle(_fromUtf8(""))
        self._tail.setObjectName(_fromUtf8("Tail"))
        self.gridLayout.addWidget(self._splitter, 0, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        Form.setWindowTitle(QApplication.translate("QRangeSlider", "QRangeSlider", None))


class RangeSliderElement(QGroupBox):
    def __init__(self, parent, main, vertical=False):
        super(RangeSliderElement, self).__init__(parent)
        self.main = main
        self.vertical = vertical

    def setStyleSheet(self, style):
        """redirect style to parent groupbox"""
        self.parent().setStyleSheet(style)

    def textColor(self):
        """text paint color"""
        # return getattr(self, '__textColor', QtGui.QColor(125, 125, 125))
        return getattr(self, "__textColor", QtGui.QColor(255, 255, 255))

    def setTextColor(self, color):
        """set the text paint color"""
        if isinstance(color, tuple) and len(color) == 3:
            color = QtGui.QColor(color[0], color[1], color[2])
        elif isinstance(color, int):
            color = QtGui.QColor(color, color, color)
        setattr(self, "__textColor", color)

    def paintEvent(self, event):
        """overrides paint event to handle text"""
        # o = QStyleOption()
        # o.initFrom(self)

        qp = QtGui.QPainter()
        qp.begin(self)
        # self.style().drawPrimitive(QStyle.PE_Widget, o, qp, self)
        if self.main.drawValues():
            self.drawText(event, qp)
        qp.end()

    def calculate_real_value(self, int_value=0):
        """
        this formula uses the int value (between 0 and 100) and return the real value using
        the real_min and real_max defined and using.

        Parameters
        ----------
        int_value

        Returns
        -------

        """
        term1 = (int_value - self.main.min()) / (self.main.max() - self.main.min())
        term2 = term1 * (self.main.real_max - self.main.real_min)
        return f"{term2 + self.main.real_min: 0.5f}"


class Head(RangeSliderElement):
    """area before the handle"""

    def __init__(self, parent, main, vertical=False):
        super(Head, self).__init__(parent, main, vertical)

    def drawText(self, event, qp):
        # qp.setPen(self.textColor())
        qp.setPen(QtGui.QColor(0, 0, 0))
        qp.setFont(QtGui.QFont("Arial", 15))
        if self.main.min_at_the_bottom:
            value = self.main.max() - self.main.min()
        else:
            value = self.main.min()
        str_value = self.calculate_real_value(int_value=value)
        qp.drawText(event.rect(), QtCore.Qt.AlignLeft, str_value)


class Tail(RangeSliderElement):
    """area after the handle"""

    def __init__(self, parent, main, vertical=False):
        super(Tail, self).__init__(parent, main, vertical)

    def drawText(self, event, qp):
        qp.setPen(self.textColor())
        qp.setFont(QtGui.QFont("Arial", 15))
        if self.main.min_at_the_bottom:
            value = self.main.min()
        else:
            value = self.main.max()
        str_value = self.calculate_real_value(int_value=value)
        qp.drawText(event.rect(), QtCore.Qt.AlignRight | QtCore.Qt.AlignBottom, str_value)


class Handle(RangeSliderElement):
    """handle area"""

    def __init__(self, parent, main, vertical=False, min_value_at_the_bottom=False):
        super(Handle, self).__init__(parent, main, vertical)

    def drawText(self, event, qp):
        # qp.setPen(self.textColor())
        qp.setPen(QtGui.QColor(255, 255, 255))
        qp.setFont(QtGui.QFont("Arial", 15))

        if self.main.min_at_the_bottom:
            value = self.main.max() - self.main.start()
        else:
            value = self.main.start()
        str_value = self.calculate_real_value(int_value=value)
        qp.drawText(event.rect(), QtCore.Qt.AlignLeft, str_value)

        if self.main.min_at_the_bottom:
            value = self.main.max() - self.main.end()
        else:
            value = self.main.end()
        str_value = self.calculate_real_value(int_value=value)
        qp.drawText(event.rect(), QtCore.Qt.AlignRight | QtCore.Qt.AlignBottom, str_value)

    def mouseReleaseEvent(self, event):
        setattr(self, "__mx", None)

    def mouseMoveEvent(self, event):
        # When the range is dragged
        event.accept()
        if self.vertical:
            mx = event.globalY()
        else:
            mx = event.globalX()

        # last value
        _mx = getattr(self, "__mx", None)

        if not _mx:
            setattr(self, "__mx", mx)
            dx = 0
        else:
            dx = mx - _mx

        if self.vertical:
            pRange = self.main.height() - self.main._splitter.handleWidth()
        else:
            pRange = self.main.width() - self.main._splitter.handleWidth()
        vRange = self.main.max() - self.main.min()
        if dx < 0:
            dx = -(-dx * vRange // pRange)
        else:
            dx = dx * vRange // pRange

        # if the motion is too small,
        if dx == 0:
            event.ignore()
            return
        # otherwise, update the mouse position.
        setattr(self, "__mx", mx)

        s0 = self.main.start()
        e0 = self.main.end()
        r0 = e0 - s0
        s = s0 + dx
        e = e0 + dx
        # Do not collapse the handle
        if s < self.main.min():
            s = self.main.min()
            e = s + r0
        elif self.main.max() < e:
            e = self.main.max()
            s = e - r0

        self.main.setRange(s, e)


class QRangeSlider(QWidget, RangeSliderForm):
    """
    The QRangeSlider class implements a horizontal range slider widget.

    Inherits QWidget.

    Methods

        * __init__ (self, QWidget parent = None)
        * bool drawValues (self)
        * int end (self)
        * (int, int) getRange (self)
        * int max (self)
        * int min (self)
        * int start (self)
        * setBackgroundStyle (self, QString styleSheet)
        * setDrawValues (self, bool draw)
        * setEnd (self, int end)
        * setStart (self, int start)
        * setRange (self, int start, int end)
        * setSpanStyle (self, QString styleSheet)

    Signals

        * endValueChanged (int)
        * maxValueChanged (int)
        * minValueChanged (int)
        * startValueChanged (int)

    Customizing QRangeSlider

    You can style the range slider as below:
    ::
        QRangeSlider * {
            border: 0px;
            padding: 0px;
        }
        QRangeSlider #Head {
            background: #222;
        }
        QRangeSlider #Span {
            background: #393;
        }
        QRangeSlider #Span:active {
            background: #282;
        }
        QRangeSlider #Tail {
            background: #222;
        }

    Styling the range slider handles follows QSplitter options:
    ::
        QRangeSlider > QSplitter::handle {
            background: #393;
        }
        QRangeSlider > QSplitter::handle:vertical {
            height: 4px;
        }
        QRangeSlider > QSplitter::handle:pressed {
            background: #ca5;
        }

    """

    # endValueChanged = Signal(int)
    # maxValueChanged = Signal(int)
    # minValueChanged = Signal(int)
    # startValueChanged = Signal(int)

    # define splitter indices
    _SPLIT_START = 1
    _SPLIT_END = 2

    # signals
    minValueChanged = Signal(int)
    maxValueChanged = Signal(int)
    startValueChanged = Signal(int)
    endValueChanged = Signal(int)

    real_end = None
    real_start = None

    def __init__(
        self,
        parent=None,
        min_value=0,
        max_value=99,
        start_value=0,
        end_value=99,
        splitterWidth=4,
        vertical=False,
        min_at_the_bottom=False,
    ):
        """Create a new QRangeSlider instance.

        :param parent: QWidget parent
        :param splitterWidth: int
        :param vertical: boolean (False)
        :param min_at_the_bottom: boolean (False) min value at the bottom when True
        :return: New QRangeSlider instance.

        """
        super(QRangeSlider, self).__init__(parent)
        self.vertical = vertical
        self.min_at_the_bottom = min_at_the_bottom
        self.setupUi(self, splitterWidth=splitterWidth, vertical=self.vertical)
        self.setMouseTracking(False)

        # self._splitter.setChildrenCollapsible(False)
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)

        # head layout
        self._head_layout = QHBoxLayout()
        self._head_layout.setSpacing(0)
        self._head_layout.setContentsMargins(0, 0, 0, 0)
        self._head.setLayout(self._head_layout)
        self.head = Head(self._head, main=self, vertical=self.vertical)
        self._head_layout.addWidget(self.head)

        # handle layout
        self._handle_layout = QHBoxLayout()
        self._handle_layout.setSpacing(0)
        self._handle_layout.setContentsMargins(0, 0, 0, 0)
        self._handle.setLayout(self._handle_layout)
        self.handle = Handle(self._handle, main=self, vertical=self.vertical)
        self.handle.setTextColor((0, 0, 0))
        self._handle_layout.addWidget(self.handle)

        # tail layout
        self._tail_layout = QHBoxLayout()
        self._tail_layout.setSpacing(0)
        self._tail_layout.setContentsMargins(0, 0, 0, 0)
        self._tail.setLayout(self._tail_layout)
        self.tail = Tail(self._tail, main=self, vertical=self.vertical)
        self._tail_layout.addWidget(self.tail)

        # defaults
        self.setMin(min_value)
        self.setMax(max_value)
        self._setMinimumRange(0)
        self.setStart(start_value)
        self.setEnd(end_value)
        self.setDrawValues(True)

    def setRealMin(self, value):
        self.real_min = value

    def setRealMax(self, value):
        self.real_max = value

    def min(self):
        """:return: minimum value"""
        return getattr(self, "__min", None)

    def max(self):
        """:return: maximum value"""
        return getattr(self, "__max", None)

    def setMin(self, value):
        """sets minimum value"""
        assert type(value) is int
        setattr(self, "__min", value)
        self.range_slider_min = value
        self.minValueChanged.emit(value)

    def setMax(self, value):
        """sets maximum value"""
        assert type(value) is int
        setattr(self, "__max", value)
        self.range_slider_max = value
        self.maxValueChanged.emit(value)

    def start(self):
        """:return: range slider start value"""
        return getattr(self, "__start", None)

    def end(self):
        """:return: range slider end value"""
        return getattr(self, "__end", None)

    def minimumRange(self):
        """:return: range slider minimum width"""
        return getattr(self, "__minimumRange", None)

    def _setStart(self, value):
        """stores the start value only"""
        setattr(self, "__start", value)
        self.startValueChanged.emit(value)

    def setStart(self, value):
        """sets the range slider start value"""
        assert type(value) is int
        if value < self.min():
            value = self.min()
        rightEnd = self.max()
        if self.end() is not None and self.end() < rightEnd:
            rightEnd = self.end()
        rightEnd -= self.minimumRange()
        if rightEnd < value:
            value = rightEnd
        v = self._valueToPos(value)
        self._splitter.splitterMoved.disconnect()
        self._splitter.moveSplitter(v, self._SPLIT_START)
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)
        self._setStart(value)

    def _setEnd(self, value):
        """stores the end value only"""
        setattr(self, "__end", value)
        self.endValueChanged.emit(value)

    def setEnd(self, value):
        """set the range slider end value"""
        assert type(value) is int
        if self.max() < value:
            value = self.max()
        leftEnd = max(self.min(), self.start()) + self.minimumRange()
        if value < leftEnd:
            value = leftEnd
        v = self._valueToPos(value) + self._splitter.handleWidth()
        self._splitter.splitterMoved.disconnect()
        self._splitter.moveSplitter(v, self._SPLIT_END)
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)
        self._setEnd(value)

    def _setMinimumRange(self, value):
        """stores the minimum range only"""
        setattr(self, "__minimumRange", value)

    def setMinimumRange(self, value):
        assert type(value) is int
        if self.max() - self.min() < value:
            value = self.max() - self.min()
        d = self.end() - self.start()
        if d < value:
            dh = d // 2
            self.setStart(self.start() - dh)
            self.setEnd(self.end() + (d - dh))
        setattr(self, "__minimumRange", value)

    def drawValues(self):
        """:return: True if slider values will be drawn"""
        return getattr(self, "__drawValues", None)

    def setDrawValues(self, draw):
        """sets draw values boolean to draw slider values"""
        assert type(draw) is bool
        setattr(self, "__drawValues", draw)

    def getRange(self):
        """:return: the start and end values as a tuple"""
        return (self.start(), self.end())

    def setRange(self, start, end, minimumRange=None):
        """set the start and end values"""
        self._setStart(start)
        self._setEnd(end)
        self.setStart(start)
        self.setEnd(end)
        if minimumRange is not None:
            self.setMinimumRange(minimumRange)

    def get_real_value_from_slider_value(self, value):
        slider_min = self.range_slider_min
        slider_max = self.range_slider_max

        if self.real_min is None:
            return

        if self.real_max is None:
            return

        coeff = (value - slider_min) / (slider_max - slider_min)
        return coeff * (self.real_min - self.real_max) + self.real_max

    def setRealRange(self, realStart, realEnd):
        real_min = self.real_min
        real_max = self.real_max
        min = self.min()
        max = self.max()

        self.real_end = realEnd
        self.real_start = realStart

        # print(f"in setRealRange")
        # print(f"{real_min =}, {real_max =}")
        # print(f"{min =}, {max =}")
        # print(f"{realStart =}, {realEnd =}")

        def get_coeff(real_value):
            coeff = (real_value - real_max) / (real_min - real_max)
            return coeff * (max - min)

        _end = int(get_coeff(realEnd) + min)
        _start = int(get_coeff(realStart) + min)

        # print(f"{_end =}")
        # print(f"{_start =}")

        self.setRange(_end, _start)

    def setRealRangeMax(self, value):
        real_start = self.real_start
        self.setRealRange(real_start, value)

    def setRealRangeMin(self, value):
        real_end = self.real_end
        self.setRealRange(value, real_end)

    def keyPressEvent(self, event):
        """overrides key press event to move range left and right"""

        key = event.key()
        if key in (QtCore.Qt.Key_Up, QtCore.Qt.Key_Left):
            s = self.start() - 1
            e = self.end() - 1
        elif key in (QtCore.Qt.Key_Down, QtCore.Qt.Key_Right):
            s = self.start() + 1
            e = self.end() + 1
        else:
            event.ignore()
            return
        event.accept()
        if s >= self.min() and e <= self.max():
            self.setRange(s, e)

    def setBackgroundStyle(self, style):
        """sets background style"""
        self._tail.setStyleSheet(style)
        self._head.setStyleSheet(style)

    def setSpanStyle(self, style):
        """sets range span handle style"""
        self._handle.setStyleSheet(style)

    def _valueToPos(self, value):
        """converts slider value to local pixel x coord"""
        # consider the splitter width
        if self.vertical:
            return scale(
                value,
                (self.min(), self.max()),
                (0, self.height() - self._splitter.handleWidth() * 2),
            )
        else:
            return scale(
                value,
                (self.min(), self.max()),
                (0, self.width() - self._splitter.handleWidth() * 2),
            )

    def _posToValue(self, pos):
        """converts local pixel x coord to slider value"""
        # consider the splitter width
        if self.vertical:
            return scale(
                pos,
                (0, self.height() - self._splitter.handleWidth() * 2),
                (self.min(), self.max()),
            )
        else:
            return scale(
                pos,
                (0, self.width() - self._splitter.handleWidth() * 2),
                (self.min(), self.max()),
            )

    def _handleMoveSplitter(self, pos, index):
        """private method for handling moving splitter handles"""
        hw = self._splitter.handleWidth()
        if index == self._SPLIT_START:
            v = self._posToValue(pos)
            # _lockPos(self._tail)
            if v + self.minimumRange() >= self.end():
                self.setEnd(v + self.minimumRange())
                v = self.end() - self.minimumRange()
            self.setStart(v)

        elif index == self._SPLIT_END:
            v = self._posToValue(pos - hw)
            # _lockPos(self._head)
            if v - self.minimumRange() <= self.start():
                self.setStart(v - self.minimumRange())
                v = self.start() + self.minimumRange()
            self.setEnd(v)


class FakeKey:
    _key = None

    def __init__(self, key="up"):
        if key == "up":
            self._key = QtCore.Qt.Key_Up
        elif key == "down":
            self._key = QtCore.Qt.Key_Down
        elif key == "left":
            self._key = QtCore.Qt.Key_Left
        else:
            self._key = QtCore.Qt.Key_Right

    def key(self):
        return self._key

    def ignore(self):
        return False

    def accept(self):
        return True


# -------------------------------------------------------------------------------
# MAIN
# -------------------------------------------------------------------------------
if __name__ == "__main__":
    app = QApplication(sys.argv)
    rs = QRangeSlider(splitterWidth=20, vertical=True, min_at_the_bottom=True)
    rs.setMin(0)
    rs.setMax(1000)
    rs.setRealMax(15.0)
    rs.setRealMin(1.0)
    # app.processEvents()
    rs.setRealRange(5, 10)  # mandatory
    # rs.setRealRangeMin(3)
    rs.setBackgroundStyle("background: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #ddd, stop:1 #333);")
    rs.handle.setStyleSheet("background: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #282, stop:1 #222);")
    rs.show()
    my_fake_key = FakeKey()
    rs.keyPressEvent(my_fake_key)
    app.exec_()



#!/usr/bin/env python
"""
ExportFittingHandler class for handling the export of the fitting parameters.
"""

import os
import shutil

import numpy as np
from qtpy import QtCore
from qtpy.QtWidgets import QApplication, QFileDialog

from ibeatles.fitting.fitting_functions import advanced_fit, basic_fit
from ibeatles.utilities.file_handler import FileHandler


class ExportFittingHandler(object):
    table: list = []
    x_axis: list = []

    def __init__(self, grand_parent=None):
        self.grand_parent = grand_parent

    def run(self):
        output_folder_name = self.select_and_create_output_folder()
        if output_folder_name == "":
            return

        # hourglass cursor
        QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        # load table
        self.load_table()

        # retrieve x-axis range selected
        x_axis = self.retrieve_x_axis()

        # loop over all bins
        for _key in self.table:
            _entry = self.table[_key]

            # create file name
            row_index = _entry["row_index"]
            column_index = _entry["column_index"]
            full_entry_file_name = self.build_entry_file_name(
                output_folder_name=output_folder_name,
                row_index=row_index,
                column_index=column_index,
            )

            # retrieve exp y_axis
            exp_y_axis = self.retrieve_y_axis(_entry)

            # retrieve fitted y_axis
            d_spacing = _entry["d_spacing"]["val"]
            sigma = _entry["sigma"]["val"]
            alpha = _entry["alpha"]["val"]
            a1 = _entry["a1"]["val"]
            a2 = _entry["a2"]["val"]

            if self.grand_parent.fitting_ui.ui.advanced_table_checkBox.isChecked():
                a5 = _entry["a5"]["val"]
                a6 = _entry["a6"]["val"]
            else:
                a5 = "N/A"
                a6 = "N/A"

            fit_y_axis = []
            for _x in x_axis:
                if self.grand_parent.fitting_ui.ui.advanced_table_checkBox.isChecked():
                    _y = advanced_fit(_x, d_spacing, alpha, sigma, a1, a2, a5, a6)
                else:
                    _y = basic_fit(_x, d_spacing, alpha, sigma, a1, a2)
                fit_y_axis.append(_y)

            # export file
            metadata = ["# bin parameters"]
            metadata.append("#   x0: {}".format(self.xy_parameters["x0"]))
            metadata.append("#   x1: {}".format(self.xy_parameters["x1"]))
            metadata.append("#   y0: {}".format(self.xy_parameters["y0"]))
            metadata.append("#   y1: {}".format(self.xy_parameters["y1"]))
            metadata.append("# fitting parameters")
            metadata.append("#   d_spacing: {}".format(d_spacing))
            metadata.append("#   sigma: {}".format(sigma))
            metadata.append("#   alpha: {}".format(alpha))
            metadata.append("#   a1: {}".format(a1))
            metadata.append("#   a2: {}".format(a2))
            metadata.append("#   a5: {}".format(a5))
            metadata.append("#   a6: {}".format(a6))
            metadata.append("#")
            metadata.append("# lambda(Angstroms), experimental, fitting")

            data = list(zip(x_axis, exp_y_axis, fit_y_axis))

            FileHandler.make_ascii_file(metadata=metadata, data=data, output_file_name=full_entry_file_name)

        QApplication.restoreOverrideCursor()

    def build_entry_file_name(self, output_folder_name="", row_index=0, column_index=0):
        """built the entry (bin number) file name"""
        return os.path.join(
            output_folder_name,
            "bin_row" + str(row_index) + "_col" + str(column_index) + ".txt",
        )

    def retrieve_y_axis(self, _entry):
        x0 = _entry["bin_coordinates"]["x0"]
        x1 = _entry["bin_coordinates"]["x1"]
        y0 = _entry["bin_coordinates"]["y0"]
        y1 = _entry["bin_coordinates"]["y1"]

        self.xy_parameters = {"x0": x0, "y0": y0, "x1": x1, "y1": y1}

        data_2d = np.array(self.grand_parent.data_metadata["normalized"]["data"])
        _data = data_2d[:, x0:x1, y0:y1]
        inter1 = np.nanmean(_data, axis=1)
        bragg_edge = np.nanmean(inter1, axis=1)

        [left_index, right_index] = self.grand_parent.fitting_bragg_edge_linear_selection

        return bragg_edge[left_index:right_index]

    def retrieve_x_axis(self):
        # index of selection in bragg edge plot
        [left_index, right_index] = self.grand_parent.fitting_bragg_edge_linear_selection

        # retrieve image
        # data_2d = np.array(self.grand_parent.data_metadata['normalized']['data'])
        full_x_axis = self.grand_parent.fitting_ui.bragg_edge_data["x_axis"]
        x_axis = np.array(full_x_axis[left_index:right_index], dtype=float)

        return x_axis

    def load_table(self):
        self.table = self.grand_parent.march_table_dictionary

    def select_and_create_output_folder(self):
        """select where to create the output fitting folder"""
        output_folder = self.grand_parent.normalized_folder
        new_output_folder = str(
            QFileDialog.getExistingDirectory(
                self.grand_parent,
                "Select Where to Create all the Fitted Bin Files....",
                output_folder,
            )
        )
        if new_output_folder:
            # define name of output folder
            default_folder_name = str(self.grand_parent.ui.normalized_folder.text()) + "_bins_fit"
            full_folder_name = os.path.join(new_output_folder, default_folder_name)
            if os.path.exists(full_folder_name):
                shutil.rmtree(full_folder_name)
            else:
                os.mkdir(full_folder_name)

            return new_output_folder

        return ""
